# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import imghdr

from flask import request, url_for
from pathlib import Path

from abilian.core.models.subjects import User
from abilian.testing import BaseTestCase
from abilian.web.preferences.user import UserPreferencesForm

AVATAR_COLORMAP = Path(__file__).parent / u'avatar-colormap.png'


class TestUserPreferences(BaseTestCase):
  SERVICES = ('security',)

  def setUp(self):
    BaseTestCase.setUp(self)
    self.user = User(
      email=u'john@example.com',
      first_name=u'John', last_name=u'Doe', can_login=True)
    self.session.add(self.user)
    self.session.commit()

  def test_form_photo(self):
    url = url_for('preferences.user')
    uploads = self.app.extensions['uploads']
    with AVATAR_COLORMAP.open('rb') as f:
      handle = uploads.add_file(self.user, f,
                                filename=u'avatar.png',
                                mimetype='image/png')

    kwargs = dict(
        method='POST',
        data={'photo': handle,},
      )

    with self.app.test_request_context(url, **kwargs):
      self.login(self.user)
      form = UserPreferencesForm(request.form)
      form.validate()
      assert form.photo.data is not None
      img_type = imghdr.what('ignored', form.photo.data)
      assert img_type == 'jpeg'



"""
Provides tools (currently: only functions, not a real service) for image
processing.
"""
from __future__ import absolute_import, division, print_function

import hashlib
from cStringIO import StringIO

from future.utils import string_types
from PIL import Image

__all__ = ['resize', 'RESIZE_MODES', 'SCALE', 'FIT', 'CROP']

# resize modes

#: resize without retaining original proportions
SCALE = 'scale'

#: resize image and retain original proportions. Image width and height will be
#: at most specified width and height, respectively; At least width or height
#: will be equal to specified width and height, respectively.
FIT = 'fit'

#: crop image and resize so that it matches specified width and height.
CROP = 'crop'

RESIZE_MODES = frozenset({SCALE, FIT, CROP})


# TODO: cache to file
cache = {}


def open_image(img):
  if isinstance(img, string_types):
    img = StringIO(img)

  pos = img.tell()
  image = Image.open(img)
  img.seek(pos)
  return image


def get_format(img):
  image = open_image(img)
  return image.format


def get_size(img):
  image = open_image(img)
  return image.size


def get_save_format(fmt):
  if fmt in ('GIF', 'PNG'):
    return 'PNG'
  return 'JPEG'


def resize(orig, width, height, mode=FIT):
  """
  """
  if isinstance(orig, string_types):
    orig = StringIO(orig)

  digest = hashlib.md5(orig.read()).digest()
  cache_key = (digest, mode, width, height)
  if cache_key in cache:
    return cache[cache_key]

  orig.seek(0)
  image = open_image(orig)
  format = image.format
  x, y = image.size

  if (x, y) == (width, height):
    orig.seek(0)
    return orig.read()

  if mode is SCALE:
    image = image.resize((width, height), Image.LANCZOS)
    assert image.size == (width, height)
  elif mode is FIT:
    if (x >= width or y >= height):
      # resize only if images exceed desired dimensions
      image.thumbnail((width, height), Image.LANCZOS)
      x1, y1 = image.size
      assert x1 == width or y1 == height
  elif mode is CROP:
    image = _crop_and_resize(image, width, height)
    assert image.size == (width, height)

  output = StringIO()
  image.save(output, get_save_format(format))
  converted = output.getvalue()
  cache[cache_key] = converted
  return converted


def _crop_and_resize(image, width, height=0):
  if not height:
    height = width

  # Compute cropping coordinates
  x0, y0 = image.size

  w_ratio = 1.0 * x0 / width
  h_ratio = 1.0 * y0 / height

  if h_ratio > w_ratio:
    x1 = 0
    x2 = x0
    y1 = int((y0 - x0 * height / width) / 2)
    y2 = int((y0 + x0 * height / width) / 2)
  else:
    x1 = int((x0 - y0 * width / height) / 2)
    x2 = int((x0 + y0 * width / height) / 2)
    y1 = 0
    y2 = y0

  image = image.crop((x1, y1, x2, y2))
  # image.load()
  image = image.resize((width, height), Image.LANCZOS)
  return image



# coding=utf-8
"""
Blob. References to files stored in a on-disk repository
"""
from __future__ import absolute_import, division, print_function

import hashlib
import uuid

import sqlalchemy as sa
from flask_sqlalchemy import BaseQuery
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer

from abilian.core.models.base import Model
from abilian.core.sqlalchemy import UUID, JSONDict


class BlobQuery(BaseQuery):
  """
  Query class for :class:`Blob` objects
  """
  def by_uuid(self, uuid):
    """
    Like `.get()` but by uuid

    :param uuid: a `string` or an `uuid`.
    """
    try:
      return self.filter_by(uuid=uuid).one()
    except sa.orm.exc.NoResultFound:
      return None


class Blob(Model):
  """
  Model for storing large file content.

  Files are stored on-disk, named after their uuid. Repository is located in
  instance folder/data/files.
  """
  __tablename__ = "blob"
  query_class = BlobQuery

  id = Column(Integer(), primary_key=True, autoincrement=True)
  uuid = Column(UUID(), unique=True, nullable=False, default=uuid.uuid4)
  meta = Column(JSONDict(), nullable=False, default=dict)

  def __init__(self, value=None, *args, **kwargs):
    super(Blob, self).__init__(*args, **kwargs)
    if self.uuid is None:
      self.uuid = uuid.uuid4()

    if self.meta is None:
      self.meta = dict()

    if value is not None:
      self.value = value

  @property
  def file(self):
    """
    Return :class:`pathlib.Path` object used for storing value
    """
    from abilian.services.repository import session_repository as repository
    return repository.get(self, self.uuid)

  @property
  def size(self):
    """
    Return size in bytes of value
    """
    f = self.file
    return f.stat().st_size if f is not None else 0

  @property
  def value(self):
    """
    Binary value content
    """
    v = self.file
    return v.open('rb').read() if v is not None else v

  @value.setter
  def value(self, value, encoding='utf-8'):
    """
    Store binary content to applications's repository and update
    `self.meta['md5']`.

    :param:content: string, bytes, or any object with a `read()` method
    :param:encoding: encoding to use when content is unicode
    """
    from abilian.services.repository import session_repository as repository
    repository.set(self, self.uuid, value)
    self.meta['md5'] = unicode(hashlib.md5(self.value).hexdigest())

    if hasattr(value, 'filename'):
      filename = getattr(value, 'filename')
      if isinstance(filename, bytes):
        filename = filename.decode('utf-8')
      self.meta['filename'] = filename

    if hasattr(value, 'content_type'):
      self.meta['mimetype'] = getattr(value, 'content_type')

  @value.deleter
  def value(self):
    """
    Remove value from repository
    """
    from abilian.services.repository import session_repository as repository
    repository.delete(self, self.uuid)


  @property
  def md5(self):
    """
    Return md5 from meta, or compute it if absent
    """
    md5 = self.meta.get('md5')
    if md5 is None:
      md5 = unicode(hashlib.md5(self.value).hexdigest())

    return md5


  def __nonzero__(self):
    """
    A blob is considered null if it has no file
    """
    return self.file is not None and self.file.exists()


@sa.event.listens_for(sa.orm.Session, 'after_flush')
def _blob_propagate_delete_content(session, flush_context):
  """
  """
  deleted = (obj for obj in session.deleted if isinstance(obj, Blob))
  for blob in deleted:
    del blob.value



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from datetime import datetime, timedelta

import pandas as pd
import sqlalchemy as sa
from flask import current_app, render_template
from numpy import sum as numpysum

from abilian.core.models.subjects import User
from abilian.i18n import _, _l
from abilian.services.audit import CREATION, AuditEntry
from abilian.services.auth.models import LoginSession

from ..panel import AdminPanel


class DashboardPanel(AdminPanel):
  id = 'dashboard'
  path = ''
  label = _l('Dashboard')
  icon = 'eye-open'

  def get(self):
    login_entries = LoginSession.query.order_by(LoginSession.started_at.asc()).all()
    # .options(sa.orm.joinedload(LoginSession.user))
    daily, weekly, monthly = uniquelogins(login_entries)
    new_logins, total_users = newlogins(login_entries)

    stats = {
      'today': stats_since(timedelta(days=1)),
      'this_week': stats_since(timedelta(days=7)),
      'this_month': stats_since(timedelta(days=30)),
    }

    # let's format the data into NVD3 datastructures
    connections = [
      {'key': _('Daily'), 'color': '#7777ff', 'values': daily},
      {'key': _('Weekly'), 'color': '#2ca02c', 'values': weekly, 'disabled': True},
      {'key': _('Monthly'), 'color': '#ff7f0e', 'values': monthly, 'disabled': True},
    ]
    new_logins = [
      {'key': _('New'), 'color': '#ff7f0e', "bar": True, 'values': new_logins},
      {'key': _('Total'), 'color': '#2ca02c', 'values': total_users}]

    return render_template("admin/dashboard.html",
                           stats=stats,
                           connections=connections,
                           new_logins=new_logins)


def stats_since(dt):
  new_members = new_documents = new_messages = 0
  after_date = datetime.utcnow() - dt
  session = current_app.db.session()
  counts_per_type = session\
    .query(AuditEntry.entity_type.label('type'),
           sa.func.count(AuditEntry.entity_type).label('count'))\
    .group_by(AuditEntry.entity_type)\
    .filter(AuditEntry.happened_at > after_date) \
    .filter(AuditEntry.type == CREATION)\
    .all()

  for entity_type, count in counts_per_type:
    entity_class = entity_type.split('.')[-1]
    if entity_class == 'User':
      new_members = count
    elif entity_class == 'Document':
      new_documents = count
    elif entity_class == 'Message':
      new_messages = count

  active_users = session.query(sa.func.count(User.id))\
                        .filter(User.last_active > after_date).scalar()

  return {
    'new_members': new_members,
    'active_users': active_users,
    'new_documents': new_documents,
    'new_messages': new_messages,
  }


epoch = datetime.utcfromtimestamp(0)


def unix_time_millis(dt):
    return (dt - epoch).total_seconds() * 1000.0


def newlogins(sessions):
  """
  brand New logins each day, and total of users each day
  :return: data, total
  2 lists of dictionaries of the following format [{'x':epoch, 'y': value},]
  """
  if not sessions:
    return [], []
  users = {}
  dates = {}

  for session in sessions:
    user = session.user
    # time value is discarded to aggregate on days only
    date = session.started_at.strftime("%Y/%m/%d")
    # keep the info only it's the first time we encounter a user
    if user not in users:
      users[user] = date
      # build the list of users on a given day
      if date not in dates:
        dates[date] = [user]
      else:
        dates[date].append(user)

  data = []
  total = []
  previous = 0
  for date in sorted(dates.keys()):
    # print u"{} : {}".format(date, len(dates[date]))
    date_epoch = unix_time_millis(datetime.strptime(date, "%Y/%m/%d"))
    data.append({u'x': date_epoch, u'y': len(dates[date])})
    previous += len(dates[date])
    total.append({u'x': date_epoch, u'y': previous})

  return data, total


def uniquelogins(sessions):
  """
  unique logins per days/weeks/months

  :return: daily, weekly, monthly
  3 lists of dictionaries of the following format [{'x':epoch, 'y': value},]
  """
  # sessions = LoginSession.query.order_by(LoginSession.started_at.asc()).all()
  if not sessions:
    return [], [], []
  dates = {}
  for session in sessions:
    user = session.user
    # time value is discarded to aggregate on days only
    date = session.started_at.strftime("%Y/%m/%d")

    if date not in dates:
      dates[date] = set()  # we want unique users on a given day
      dates[date].add(user)
    else:
      dates[date].add(user)

  daily = []
  weekly = []
  monthly = []

  for date in sorted(dates.keys()):
    # print u"{} : {}".format(date, len(dates[date]))
    date_epoch = unix_time_millis(datetime.strptime(date, "%Y/%m/%d"))
    daily.append({u'x': date_epoch, u'y': len(dates[date])})

  # first_day = data[0]['x']
  # last_day = data[-1]['x']

  daily_serie = pd.Series(dates)
  # convert the index to Datetime type
  daily_serie.index = pd.DatetimeIndex(daily_serie.index)
  # calculate the values instead of users lists
  daily_serie = daily_serie.apply(lambda x: len(x))

  # GroupBy Week/month, Thanks Panda
  weekly_serie = daily_serie.groupby(pd.TimeGrouper(freq='W')).aggregate(numpysum)
  monthly_serie = daily_serie.groupby(pd.TimeGrouper(freq='M')).aggregate(numpysum)

  for date, value in weekly_serie.iteritems():
    date_epoch = unix_time_millis(date)
    weekly.append({u'x': date_epoch, u'y': value})

  for date, value in monthly_serie.iteritems():
    date_epoch = unix_time_millis(date)
    monthly.append({u'x': date_epoch, u'y': value})

  return daily, weekly, monthly



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

import os
import io
import logging
import pathlib

logger = logging.getLogger(__name__)

try:
  import clamd
  cd = clamd.ClamdUnixSocket()
  CLAMD_AVAILABLE = True
except ImportError:
  CLAMD_AVAILABLE = False

from abilian.core.models.blob import Blob
from ..base import Service

CLAMD_CONF = {
  'StreamMaxLength': u'25M',
  'MaxFileSize': u'25M',
}
CLAMD_STREAMMAXLENGTH = 26214400
CLAMD_MAXFILESIZE = 26214400

if CLAMD_AVAILABLE:
  conf_path = pathlib.Path('/etc', 'clamav', 'clamd.conf')
  if conf_path.exists():
    conf_lines = [l.strip() for l in conf_path.open('rt').readlines()]
    CLAMD_CONF = dict(l.split(u' ', 1) for l in conf_lines
                      if not l.startswith(u'#'))

    def _size_to_int(size_str):
      multiplier = 0
      if not size_str:
        return 0

      unit = size_str[-1].lower()
      if unit in ('k', 'm'):
        size_str = size_str[:-1]
        multiplier = 1024
        if unit == 'm':
          multiplier *= 1024

      if not size_str:
        return 0

      size = int(size_str)
      if multiplier:
        size *= multiplier

      return size

    CLAMD_STREAMMAXLENGTH = _size_to_int(CLAMD_CONF['StreamMaxLength'])
    CLAMD_MAXFILESIZE = _size_to_int(CLAMD_CONF['MaxFileSize'])
    del conf_path, conf_lines, _size_to_int


class AntiVirusService(Service):
  """
  Antivirus service
  """
  name = 'antivirus'

  def scan(self, file_or_stream):
    """
    :param file_or_stream: :class:`Blob` instance, filename or file object

    :returns: True if file is 'clean', False if a virus is detected, None if
    file could not be scanned.

    If `file_or_stream` is a Blob, scan result is stored in
    Blob.meta['antivirus'].
    """
    res = self._scan(file_or_stream)
    if isinstance(file_or_stream, Blob):
      file_or_stream.meta['antivirus'] = res
    return res

  def _scan(self, file_or_stream):
    if not CLAMD_AVAILABLE:
      return None

    content = file_or_stream
    if isinstance(file_or_stream, Blob):
      # py3 compat: bytes == py2 str(). Pathlib uses os.fsencode()
      file_or_stream = bytes(file_or_stream.file)
    elif isinstance(file_or_stream, unicode):
      file_or_stream = file_or_stream.encode(os.fsencode)

    if isinstance(file_or_stream, bytes):
      content = io.open(file_or_stream, 'rb')

    if content.seekable():
      pos = content.tell()
      content.seek(0, io.SEEK_END)
      size = content.tell()
      content.seek(pos)

      if size > CLAMD_STREAMMAXLENGTH:
        logger.error(
          'Content size exceed antivirus size limit, size=%d, limit=%d (%s)',
          size,
          CLAMD_STREAMMAXLENGTH,
          CLAMD_CONF['StreamMaxLength'].encode('utf-8'),
          extra={'stack': True},
        )
        return None

    # use stream scan. When using scan by filename, clamd runnnig user must have
    # access to file, which we cannot guarantee
    scan = cd.instream
    res = None
    try:
      res = scan(content)
    except clamd.ClamdError as e:
      self.logger.warning('Error during content scan: %s', repr(e))
      return None

    if 'stream' not in res:
      # may happen if file doesn't exists
      return False

    res = res['stream']
    return res[0] == u'OK'


service = AntiVirusService()



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division


from .forms import AttachmentForm
from .extension import AttachmentExtension, AttachmentsManager

def register_plugin(app):
  AttachmentExtension(app=app)



# coding=utf-8
"""
Subject classes (i.e. people, groups, etc.).

See ICOM-ics-v1.0 "Subject Branch".

TODO: I'm not a big fan of the "subject" name. Could be replaced by something
else, like "people" or "principal" ?
"""
from __future__ import absolute_import, division, print_function

import random
import string
from abc import ABCMeta, abstractmethod, abstractproperty
from datetime import datetime, timedelta

import bcrypt
import sqlalchemy as sa
from flask_login import UserMixin, current_app
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, deferred, relationship
from sqlalchemy.schema import Column, ForeignKey, Table, UniqueConstraint
from sqlalchemy.types import (Boolean, DateTime, Integer, LargeBinary,
                              UnicodeText)

from abilian.core import sqlalchemy as sa_types

from .base import SEARCHABLE, SYSTEM, IdMixin, Indexable, TimestampedMixin, db

__all__ = ['User', 'Group', 'Principal']


# Tables for many-to-many relationships
following = Table(
  'following', db.Model.metadata,
  Column('follower_id', Integer, ForeignKey('user.id')),
  Column('followee_id', Integer, ForeignKey('user.id')),
  UniqueConstraint('follower_id', 'followee_id'),
)
membership = Table(
  'membership', db.Model.metadata,
  Column('user_id', Integer,
         ForeignKey('user.id', onupdate='CASCADE', ondelete='CASCADE')),
  Column('group_id', Integer,
         ForeignKey('group.id', onupdate='CASCADE', ondelete='CASCADE')),
  UniqueConstraint('user_id', 'group_id'),
)

# Should not be needed (?)
administratorship = Table(
  'administratorship', db.Model.metadata,
  Column('user_id', Integer,
         ForeignKey('user.id', onupdate='CASCADE', ondelete='CASCADE')),
  Column('group_id', Integer,
         ForeignKey('group.id', onupdate='CASCADE', ondelete='CASCADE')),
  UniqueConstraint('user_id', 'group_id'),
)

_RANDOM_PASSWORD_CHARS = (string.ascii_letters
                          + string.digits
                          + string.punctuation)


def gen_random_password(length=15):
  """
  """
  rg = random.SystemRandom()
  return u''.join(rg.choice(_RANDOM_PASSWORD_CHARS) for i in range(length))


class PasswordStrategy(object):
  """

  """
  __metaclass__ = ABCMeta

  @abstractproperty
  def name(self):
    """
    Strategy name.
    """

  @abstractmethod
  def authenticate(self, user, password):
    """
    Predicate to tell wether password match user's or not.
    """

  @abstractmethod
  def process(self, user, password):
    """
    Return a string to be stored as user password
    """


class ClearPasswordStrategy(PasswordStrategy):
  """
  Don't encrypt at all.

  This strategy should not ever be used elsewhere than in tests. It's useful
  in tests since a hash like bcrypt is designed to be slow.
  """
  @property
  def name(self):
    return "clear"

  def authenticate(self, user, password):
    return user.password == password

  def process(self, user, password):
    if not isinstance(password, unicode):
      password = password.decode('utf-8')
    return password


class BcryptPasswordStrategy(PasswordStrategy):
  """
  Hash passwords using bcrypt.
  """

  @property
  def name(self):
    return 'bcrypt'

  def authenticate(self, user, password):
    current_passwd = user.password
    # crypt work only on str, not unicode
    if isinstance(current_passwd, unicode):
      current_passwd = current_passwd.encode('utf-8')
    if isinstance(password, unicode):
      password = password.encode('utf-8')

    return bcrypt.hashpw(password, current_passwd) == current_passwd

  def process(self, user, password):
    if isinstance(password, unicode):
      password = password.encode('utf-8')
    return bcrypt.hashpw(password, bcrypt.gensalt()).decode('utf-8')


class UserQuery(db.Model.query_class):
  def get_by_email(self, email):
    return self.filter_by(email=email).one()


class Principal(IdMixin, TimestampedMixin, Indexable):
  """A principal is either a User or a Group."""

  __indexation_args__ = {}
  __indexation_args__.update(Indexable.__indexation_args__)
  index_to = __indexation_args__.setdefault('index_to', ())
  __indexation_args__['index_to'] += (
      ('name', ('name', 'name_prefix', 'text')),
    )
  del index_to

  def has_role(self, role):
    return current_app.services['security'].has_role(self, role)



class User(Principal, UserMixin, db.Model):
  __tablename__ = 'user'
  __editable__ = ['first_name', 'last_name', 'email', 'password']
  __exportable__ = __editable__ + ['created_at', 'updated_at', 'id']

  __password_strategy__ = BcryptPasswordStrategy()

  entity_type = u'{}.{}'.format(__module__, 'User')

  query_class = UserQuery

  # Basic information
  first_name = Column(UnicodeText, info=SEARCHABLE)
  last_name = Column(UnicodeText, info=SEARCHABLE)
  # Should we add gender, salutation ?

  # System information
  email = Column(UnicodeText, nullable=False)
  can_login = Column(Boolean, nullable=False, default=True)
  password = Column(UnicodeText, default=u"*",
                    info={'audit_hide_content': True})

  photo = deferred(Column(LargeBinary))

  last_active = Column(DateTime, info=SYSTEM)
  locale = Column(sa_types.Locale, nullable=True, default=None)
  timezone = Column(sa_types.Timezone, nullable=True, default=None)

  __table_args__ = (UniqueConstraint('email'),)

  followers = relationship("User", secondary=following,
                           primaryjoin=('User.id == following.c.follower_id'),
                           secondaryjoin=('User.id == following.c.followee_id'),
                           backref='followees')

  def __init__(self, password=None, **kwargs):
    Principal.__init__(self)
    UserMixin.__init__(self)
    db.Model.__init__(self, **kwargs)

    if self.can_login and password is not None:
      self.set_password(password)

  def authenticate(self, password):
    if self.password and self.password != "*":
      return self.__password_strategy__.authenticate(self, password)
    else:
      return False

  def set_password(self, password):
    """Encrypts and sets password."""
    self.password = self.__password_strategy__.process(self, password)

  def follow(self, followee):
    if followee == self:
      raise Exception("User can't follow self")
    self.followees.append(followee)

  def unfollow(self, followee):
    if followee == self:
      raise Exception("User can't follow self")
    i = self.followees.index(followee)
    del self.followees[i]

  def join(self, group):
    self.groups.add(group)

  def leave(self, group):
    if group in self.groups:
      self.groups.remove(group)

  #
  # Boolean properties
  #
  def is_following(self, other):
    return other in self.followees

  def is_member_of(self, group):
    return self in group.members

  def is_admin_of(self, group):
    return self in group.admins

  @property
  def is_online(self):
    if self.last_active is None:
      return False
    return datetime.utcnow() - self.last_active <= timedelta(minutes=1)

  #
  # Other properties
  #
  @property
  def name(self):
    name = u'{first_name} {last_name}'.format(first_name=self.first_name or u'',
                                              last_name=self.last_name or u'')
    return name.strip() or u'Unknown'

  @property
  def short_name(self):
    first_name = self.first_name or u''
    last_name = self.last_name[0:1] + "." if self.last_name else u''
    name = u'{} {}'.format(first_name, last_name)
    return name.strip() or u'Unknown'

  def __unicode__(self):
    return self.name

  def __repr__(self):
    cls = self.__class__
    return '<{mod}.{cls} id={id!r} email={email!r} at 0x{addr:x}>'.format(
      mod=cls.__module__, cls=cls.__name__,
      id=self.id, email=self.email, addr=id(self))


@sa.event.listens_for(User, "mapper_configured", propagate=True)
def _add_user_indexes(mapper, class_):
  # this is a functional index (indexes on a function result), we cannot define
  # it in __table_args__.
  #
  # see: https://groups.google.com/d/msg/sqlalchemy/CgSJUlelhGs/_Nj3f201hs4J
  idx = sa.schema.Index('user_unique_lowercase_email',
                        sa.sql.func.lower(class_.email),
                        unique=True)
  idx.info['engines'] = ('postgresql',)


class Group(Principal, db.Model):
  __indexable__ = False
  __tablename__ = 'group'
  __editable__ = ['name', 'description']
  __exportable__ = __editable__ + ['created_at', 'updated_at', 'id']

  entity_type = u'{}.{}'.format(__module__, 'Group')

  name = Column(UnicodeText, nullable=False, info=SEARCHABLE)
  description = Column(UnicodeText, info=SEARCHABLE)

  members = relationship(
    "User",
    collection_class=set,
    secondary=membership,
    backref=backref('groups', lazy='select', collection_class=set)
  )
  admins = relationship("User", collection_class=set,
                        secondary=administratorship)

  photo = deferred(Column(LargeBinary))

  public = Column(Boolean, default=False, nullable=False)

  @hybrid_property
  def members_count(self):
    return len(self.members)

  @members_count.expression
  def members_count(cls):
    return sa.sql.select([sa.sql.func.count(membership.c.user_id)])\
                 .where(membership.c.group_id == cls.id)\
                 .group_by(membership.c.group_id)\
                 .label('members_count')



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from datetime import datetime, timedelta

from _pytest.python import raises

from abilian.core.entities import Entity
from abilian.testing import BaseTestCase as AbilianTestCase

from ..comment import Comment, is_commentable, register


@register
class CommentableContent(Entity):
  pass


def test_commentable_interface():
  assert is_commentable(CommentableContent)

  instance = CommentableContent(name=u'test instance')
  assert not is_commentable(instance) # not in DB: no id

  instance.id = 42
  assert is_commentable(instance)
  assert not is_commentable(object)
  assert not is_commentable(object())


def test_cannot_register_non_entities():
  class Dummy(object):
    pass

  with raises(ValueError):
    register(Dummy)


class TestComment(AbilianTestCase):

  def test_default_ordering(self):
    commentable = CommentableContent(name=u'commentable objet')
    self.session.add(commentable)

    now = datetime.now()
    c1 = Comment(entity=commentable, body=u'comment #1')
    c1.created_at = now - timedelta(10)
    self.session.flush()
    c2 = Comment(entity=commentable, body=u'comment #2')
    c2.created_at = now - timedelta(5)
    self.session.flush()

    query = Comment.query.filter(Comment.entity == commentable)
    assert query.all() == [c1, c2]

    c1.created_at = c2.created_at
    c2.created_at = c1.created_at - timedelta(5)
    assert query.all() == [c2, c1]



# coding=utf-8
"""
Admin panel for tags
"""
from __future__ import absolute_import

import logging

import sqlalchemy as sa
from flask import current_app, flash, redirect, render_template, request
from sqlalchemy.sql import functions

from abilian.core.entities import Entity
from abilian.core.models.tag import Tag, entity_tag_tbl
from abilian.i18n import _, _l, _n
from abilian.services.indexing.service import index_update
from abilian.web import url_for
from abilian.web.admin import AdminPanel
from abilian.web.views import ObjectEdit
from abilian.web.views.base import View

from .forms import TagForm

logger = logging.getLogger(__name__)

_OBJ_COUNT = functions.count(entity_tag_tbl.c.entity_id).label('obj_count')


def get_entities_for_reindex(tags):
  """
  Collect entities for theses tags.
  """
  if isinstance(tags, Tag):
    tags = (tags,)

  session = current_app.db.session()
  indexing = current_app.services['indexing']
  tbl = Entity.__table__
  tag_ids = [t.id for t in tags]
  q = sa.sql.select([tbl.c.entity_type, tbl.c.id])\
            .select_from(
              tbl.join(entity_tag_tbl,
                       entity_tag_tbl.c.entity_id == tbl.c.id))\
            .where(entity_tag_tbl.c.tag_id.in_(tag_ids))

  entities = set()

  with session.no_autoflush:
    for entity_type, entity_id in session.execute(q):
      if entity_type not in indexing.adapted:
        logger.debug('%r is not indexed, skipping', entity_type)

      item = ('changed', entity_type, entity_id, ())
      entities.add(item)

  return entities


def schedule_entities_reindex(entities):
  """
  :param entities: as returned by :func:`get_entities_for_reindex`
  """
  entities = [(e[0], e[1], e[2], dict(e[3])) for e in entities]
  return index_update.apply_async(kwargs=dict(index='default', items=entities))


class NSView(View):
  """
  View a Namespace
  """
  def __init__(self, view_endpoint, *args, **kwargs):
    super(NSView, self).__init__(*args, **kwargs)
    self.__selected_tags = None
    self.view_endpoint = view_endpoint

  def prepare_args(self, args, kwargs):
    self.ns = kwargs.get('ns')
    self.form_errors = {}
    return args, kwargs

  def get(self, ns):
    tags = Tag.query\
              .filter(Tag.ns == ns)\
              .outerjoin(entity_tag_tbl, entity_tag_tbl.c.tag_id == Tag.id)\
              .add_column(_OBJ_COUNT)\
              .group_by(Tag)\
              .order_by(sa.sql.func.lower(Tag.label))

    # get a list of rows instead of (Tag, count) tuples
    tags = list(tags.session.execute(tags))
    return render_template(
      'admin/tags_ns.html',
      ns=ns,
      tags=tags,
      errors=self.form_errors,
      merge_to=request.form.get('merge_to', default='__None__', type=int),
      selected_tags=set(t.id for t in self._get_selected_tags()),
    )

  def redirect_to_view(self):
    return redirect(url_for('.tags_ns', ns=self.ns))

  def post(self, ns):
    data = request.form
    action = data.get('__action')

    if action == u'delete':
      return self.do_delete()
    elif action == u'merge':
      return self.do_merge()

    else:
      flash(_(u'Unknown action'))
      self.get(self.ns)

  def _get_selected_tags(self):
    if self.__selected_tags is None:

      tag_ids = request.form.getlist('selected', type=int)
      if not tag_ids:
        self.__selected_tags = []
      else:
        self.__selected_tags = Tag.query.filter(Tag.ns == self.ns,
                                                Tag.id.in_(tag_ids))\
                                        .all()
    return self.__selected_tags

  def do_delete(self):
    data = request.form
    confirm = data.get('confirm_delete', False, type=bool)

    if not confirm:
      flash(_(u'Please fix the error(s) below'), 'error')
      self.form_errors['confirm_delete'] = _(u'Must be checked to ensure you '
                                             u'intent to delete these tags')
      return self.get(self.ns)

    session = current_app.db.session()
    tags = self._get_selected_tags()

    if not tags:
      flash(_(u'No action performed: no tags selected'), 'warning')
      return self.redirect_to_view()

    count = len(tags)
    entities_to_reindex = get_entities_for_reindex(tags)
    success_message = _n(u'%(tag)s deleted',
                         u'%(num)d tags deleted:\n%(tags)s',
                         count,
                         tag=tags[0].label,
                         tags=u', '.join(t.label for t in tags))
    map(session.delete, tags)
    session.commit()
    flash(success_message)
    schedule_entities_reindex(entities_to_reindex)
    return self.redirect_to_view()


  def do_merge(self):
    target_id = request.form.get('merge_to', type=int)

    if not target_id:
      flash(_(u'You must select a target tag to merge to'), 'error')
      return self.get(self.ns)

    target = Tag.query.filter(Tag.ns == self.ns, Tag.id == target_id).scalar()

    if not target:
      flash(_(u'Target tag not found, no action performed'), 'error')
      return self.get(self.ns)

    merge_from = set(self._get_selected_tags())

    if target in merge_from:
      merge_from.remove(target)

    if not merge_from:
      flash(_(u'No tag selected for merging'), 'warning')
      return self.get(self.ns)

    session = current_app.db.session()
    merge_from_ids = [t.id for t in merge_from]
    tbl = entity_tag_tbl
    entities_to_reindex = get_entities_for_reindex(merge_from)
    already_tagged = sa.sql.select([tbl.c.entity_id]).where(tbl.c.tag_id == target.id)
    del_dup = tbl.delete()\
              .where(sa.sql.and_(tbl.c.tag_id.in_(merge_from_ids),
                                 tbl.c.entity_id.in_(already_tagged)))
    session.execute(del_dup)
    update = tbl.update()\
             .where(tbl.c.tag_id.in_(merge_from_ids))\
             .values(tag_id=target.id)
    session.execute(update)
    map(session.delete, merge_from)
    session.commit()
    schedule_entities_reindex(entities_to_reindex)
    return self.redirect_to_view()


class BaseTagView(object):
  """
  Mixin for tag views
  """
  Model = Tag
  Form = TagForm

  def __init__(self, *args, **kwargs):
    super(BaseTagView, self).__init__(*args, **kwargs)
    self.extension = current_app.extensions['tags']


  def prepare_args(self, args, kwargs):
    self.ns = kwargs.pop('ns')
    return super(BaseTagView, self).prepare_args(args, kwargs)

  def view_url(self):
    return url_for('.tags_ns', ns=self.ns)

  index_url = view_url


class TagEdit(BaseTagView, ObjectEdit):
  _message_success = _l(u'Tag edited')

  def after_populate_obj(self):
    session = sa.orm.object_session(self.obj)
    self.has_changes = self.obj in (session.dirty | session.deleted)
    if self.has_changes:
      # since the tag may be in pending-delete, we must collect them before
      # flush
      self._entities_to_reindex = get_entities_for_reindex(self.obj)

  def commit_success(self):
    if not (self.has_changes and self._entities_to_reindex):
      return

    schedule_entities_reindex(self._entities_to_reindex)


class TagPanel(AdminPanel):
  """
  Tags administration
  """
  id = 'tags'
  label = _l(u'Tags')
  icon = 'tags'

  def get(self):
    obj_count = \
      sa.sql.select([
        Tag.ns,
        functions.count(entity_tag_tbl.c.entity_id).label('obj_count'),
      ])\
            .select_from(Tag.__table__.join(entity_tag_tbl))\
            .group_by(Tag.ns)\
            .alias()

    ns_query = sa.sql.select(
      [Tag.ns,
       functions.count(Tag.id).label('tag_count'),
       obj_count.c.obj_count],
      from_obj=[Tag.__table__.outerjoin(obj_count, Tag.ns == obj_count.c.ns)]
    )\
                     .group_by(Tag.ns, obj_count.c.obj_count)\
                     .order_by(Tag.ns)

    session = current_app.db.session()
    namespaces = session.execute(ns_query)

    return render_template(
      'admin/tags.html',
      namespaces=namespaces,
    )

  def install_additional_rules(self, add_url_rule):
    panel_endpoint = '.' + self.id
    ns_base = '/<string:ns>/'
    add_url_rule(ns_base, endpoint='ns',
                 view_func=NSView.as_view('ns', view_endpoint=panel_endpoint))

    tag_base = ns_base + '<int:object_id>/'
    add_url_rule(tag_base, endpoint='tag_edit',
                 view_func=TagEdit.as_view('tag_edit', view_endpoint=panel_endpoint))

    add_url_rule(tag_base + 'delete', endpoint='tag_delete',
                 view_func=TagEdit.as_view('tag_delete',
                                           view_endpoint=panel_endpoint))



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from flask import current_app
from flask_login import current_user

from abilian.core.entities import Entity
from abilian.core.models import attachment as attachments
from abilian.services.security import READ, WRITE, security
from abilian.web import url_for

from .forms import AttachmentForm
from .views import bp as blueprint
from .views import UPLOAD_BUTTON

_MANAGER_ATTR = '__attachments_manager__'


class AttachmentExtension(object):
  """
  API for comments, installed as an application extension.

  It is also available in templates as `attachments`.
  """
  def __init__(self, app):
    app.extensions['attachments'] = self
    app.add_template_global(self, 'attachments')
    app.register_blueprint(blueprint)

  def manager(self, obj):
    """
    Returns the :class:`AttachmentsManager` instance for this object.
    """
    manager = getattr(obj, _MANAGER_ATTR, None)
    if manager is None:
      manager = AttachmentsManager()
      setattr(obj.__class__, _MANAGER_ATTR, manager)

    return manager

  def is_support_attachments(self, obj):
    return self.manager(obj).is_support_attachments(obj)

  def for_entity(self, obj, check_support_attachments=False):
    return self.manager(obj).for_entity(
        obj,
        check_support_attachments=check_support_attachments,
    )

  def has_attachments(self, obj):
    return self.manager(obj).has_attachments(obj)

  def count(self, obj):
    return self.manager(obj).count(obj)

  def get_form_context(self, obj):
    """
    Return a dict: form instance, action button, submit url...
    Used by macro m_attachment_form(entity)
    """
    return self.manager(obj).get_form_context(obj)



_DEFAULT_TEMPLATE = 'macros/attachment_default.html'

class AttachmentsManager(object):
  """
  Allow customization of attachments form, display macros, etc

  can be used as class decorator
  """
  Form = AttachmentForm
  macros_template = 'macros/attachment.html'

  def __init__(self, Form=AttachmentForm,
               macros_template='macros/attachment_default.html'):
    self.Form = Form
    self.macros_template = macros_template

  def __call__(self, Model):
    setattr(Model, _MANAGER_ATTR, self)
    return Model

  @property
  def macros(self):
    default_template = current_app.jinja_env.get_template(_DEFAULT_TEMPLATE)
    template = current_app.jinja_env.get_template(self.macros_template)
    default = default_template.module
    m = template.module
    return dict(
      m_attachments=getattr(m, 'm_attachments', default.m_attachments),
      m_attachment=getattr(m, 'm_attachment', default.m_attachment),
      m_attachment_form=getattr(m, 'm_attachment_form',
                                default.m_attachment_form)
    )

  def is_support_attachments(self, obj):
    return attachments.is_support_attachments(obj)

  def for_entity(self, obj, check_support_attachments=False):
    return attachments.for_entity(obj, check_support_attachments=check_support_attachments)

  def has_attachments(self, obj):
    return bool(self.for_entity(obj, check_support_attachments=True))

  def count(self, obj):
    return len(self.for_entity(obj, check_support_attachments=True))

  def get_form_context(self, obj):
    """
    Return a dict: form instance, action button, submit url...
    Used by macro m_attachment_form(entity)
    """
    ctx = {}
    ctx['url'] = url_for('attachments.create', entity_id=obj.id)
    ctx['form'] = self.Form()
    ctx['buttons'] = [UPLOAD_BUTTON]
    return ctx

  ## current user capabilities
  def can_view(self, entity):
    """
    True if user can view attachments on entity
    """
    return security.has_permission(current_user, READ, obj=entity)

  def can_edit(self, entity):
    """
    True if user can edit attachments on entity
    """
    return security.has_permission(current_user, WRITE, obj=entity)

  def can_create(self, entity):
    """
    True if user can add attachments
    """
    return security.has_permission(current_user, WRITE, obj=entity)

  def can_delete(self, entity):
    """
    True if user can delete attachments
    """
    return security.has_permission(current_user, WRITE, obj=entity)



"""
Useful decorators for web views.
"""
from __future__ import absolute_import, division, print_function

import warnings
from functools import wraps

from flask import render_template, request

__all__ = ['templated']


# Copy/pasted from: http://flask.pocoo.org/docs/patterns/viewdecorators/
def templated(template=None):
  """
  The idea of this decorator is that you return a dictionary with the values
  passed to the template from the view function and the template
  is automatically rendered.

  @deprecated
  """
  def decorator(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
      template_name = template
      if template_name is None:
        template_name = request.endpoint.replace('.', '/') + '.html'
      ctx = f(*args, **kwargs)
      if ctx is None:
        ctx = {}
      elif not isinstance(ctx, dict):
        return ctx
      return render_template(template_name, **ctx)
    return decorated_function
  return decorator


# Copy/pasted from:
# https://wiki.python.org/moin/PythonDecoratorLibrary#Generating_Deprecation_Warnings
def deprecated(func):
    '''This decorator can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.'''

    @wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            filename=func.func_code.co_filename,
            lineno=func.func_code.co_firstlineno + 1
        )
        return func(*args, **kwargs)
    return new_func



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import uuid
from StringIO import StringIO

from abilian.services import repository_service as repository
from abilian.services import session_repository_service as session_repository
from abilian.testing import BaseTestCase as AbilianTestCase

from ..blob import Blob


#
# Unit tests
#
def test_auto_uuid():
  b = Blob()
  assert b.uuid is not None
  assert isinstance(b.uuid, uuid.UUID)

  # test provided uuid is not replaced by a new one
  u = uuid.UUID('4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9')
  b = Blob(uuid=u)
  assert isinstance(b.uuid, uuid.UUID)
  assert b.uuid, u


def test_meta():
  b = Blob()
  assert b.meta == dict()

#
# Integration tests
#
class BlobTestCase(AbilianTestCase):

  def test_md5(self):
    b = Blob(u'test md5')
    assert 'md5' in b.meta
    assert b.meta['md5'] == u'0e4e3b2681e8931c067a23c583c878d5'

  def test_size(self):
    b = Blob(u'test')
    assert b.size == 4

  def test_filename(self):
    content = StringIO(u'test')
    content.filename = 'test.txt'
    b = Blob(content)
    assert 'filename' in b.meta
    assert b.meta['filename'] == u'test.txt'

  def test_mimetype(self):
    content = StringIO(u'test')
    content.content_type = 'text/plain'
    b = Blob(content)
    assert 'mimetype' in b.meta
    assert b.meta['mimetype'] == u'text/plain'

  def test_nonzero(self):
    b = Blob(u'test md5')
    assert bool(b)

    # change uuid: repository will return None for blob.file
    b.uuid = uuid.uuid4()
    assert not bool(b)

  def test_query(self):
    session = self.session
    content = b'content'
    b = Blob(content)
    session.add(b)
    session.flush()

    assert Blob.query.by_uuid(b.uuid) is b
    assert Blob.query.by_uuid(unicode(b.uuid)) is b

    u = uuid.uuid4()
    assert Blob.query.by_uuid(u) is None

  def test_value(self):
    session = self.session
    content = b'content'
    b = Blob(content)

    tr = session.begin(nested=True)
    session.add(b)
    tr.commit()

    assert repository.get(b.uuid) is None
    assert session_repository.get(b, b.uuid).open('rb').read() == content
    assert b.value == content

    session.commit()
    assert repository.get(b.uuid).open('rb').read() == content
    assert b.value == content

    session.begin(nested=True) # match session.rollback

    with session.begin(nested=True):
      session.delete(b)
      # object marked for deletion, but instance attribute should still be
      # readable
      assert session_repository.get(b, b.uuid).open('rb').read() == content

    # commit in transaction: session_repository has no content, 'physical'
    # repository still has content
    assert session_repository.get(b, b.uuid) is None
    assert repository.get(b.uuid).open('rb').read() == content

    # rollback: session_repository has content again
    session.rollback()
    assert session_repository.get(b, b.uuid).open('rb').read() == content

    session.delete(b)
    session.flush()
    assert session_repository.get(b, b.uuid) is None
    assert repository.get(b.uuid).open('rb').read() == content

    session.commit()
    assert repository.get(b.uuid) is None



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from flask import current_app
from wtforms.fields import StringField

from abilian.core.models.tag import Tag
from abilian.web.forms import Field, Form
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required
from abilian.web.forms.widgets import ListWidget, Select2


class TagsField(Field):
  """
  Handle tags selection on for a given :attr:`tags namespace<.Tag.ns>`.

  Usage::

      __tags__ = TagsField(ns='tags namespace')

  """
  multiple = True
  widget = Select2(js_init='tags-select', multiple=True)
  view_widget = ListWidget()

  def __init__(self, ns, *args, **kwargs):
    kwargs.setdefault('view_widget', self.view_widget)
    super(TagsField, self).__init__(*args, **kwargs)
    self.ns = ns.strip()
    assert self.ns

  def iter_choices(self):
    choices = []
    extension = current_app.extensions['tags']
    ns_tags = extension.get(ns=self.ns)

    for tag in ns_tags:
      choices.append((tag.label, tag.label, tag in self.data))

    # in case of error during form validation, don't forget new added tags
    for tag in self.data - set(ns_tags):
      choices.append((tag.label, tag.label, True))

    return sorted(choices)


  def default(self):
    return set()

  def process_data(self, data):
    data = {t for t in data if t.ns == self.ns}
    return super(TagsField, self).process_data(data)

  def process_formdata(self, valuelist):
    extension = current_app.extensions['tags']
    valuelist = set(valuelist[0].split(u';'))
    data = set()

    for label in valuelist:
      label = label.strip()
      if not label:
        continue

      tag = extension.get(ns=self.ns, label=label)
      if tag is None:
        tag = Tag(ns=self.ns, label=label)
      data.add(tag)

    self.data = data

  def populate_obj(self, obj, name):
    extension = current_app.extensions['tags']
    # all_tags is an InstrumentedSet. add/remove will result in DB operations.
    all_tags = extension.entity_tags(obj)
    all_ns_tags = {t for t in extension.entity_tags(obj)
                   if t.ns == self.ns}
    to_remove = all_ns_tags - self.data

    for tag in to_remove:
      all_tags.remove(tag)

    for tag in self.data:
      all_tags.add(tag)


_NS = StringField(u'Namespace',
                   validators=[required()],
                   filters=(strip,),)

class TagForm(Form):
  """
  Form for a single tag
  """
  label = StringField(u'Label', filters=(strip,), validators=[required(),])


class TagNSForm(TagForm):
  """
  Allow to edit namespace.
  """
  ns = _NS



# coding=utf-8
"""
"""
from __future__ import absolute_import

from flask import current_app
from flask_debugtoolbar.panels import DebugPanel

from abilian.core.entities import Entity
from abilian.core.models.subjects import Group
from abilian.i18n import _
from abilian.web.action import actions

from .models import Anonymous


class SecurityInfoDebugPanel(DebugPanel):
  """
  A panel to display current roles and permissions for "current" object.
  """
  name = 'SecurityInfo'

  @property
  def current_obj(self):
    return actions.context.get('object')

  @property
  def has_content(self):
    obj = self.current_obj
    return (obj is not None
            and isinstance(obj, Entity)
            and obj.id is not None)

  def nav_title(self):
    return _('Security Info')

  def nav_subtitle(self):
    """Subtitle showing until title in toolbar"""
    obj = self.current_obj
    if not obj:
      return _(u'No current object')

    try:
      return u'{}(id={})'.format(obj.__class__.__name__, obj.id)
    except:
      return u''

  def title(self):
    return self.nav_title()

  def url(self):
      return ''

  def content(self):
    obj = self.current_obj
    svc = current_app.services['security']
    context = self.context.copy()

    context['permissions'] = svc.get_permissions_assignments(obj=obj)
    context['roles'] = roles = dict()

    for principal, r in svc.get_role_assignements(object=obj):
      if r not in roles:
        roles[r] = dict(anonymous=False, users=set(), groups=set())

      info = roles[r]
      if principal is Anonymous:
        info['anonymous'] = True
      elif isinstance(principal, Group):
        info['groups'].add(principal)
      else:
        info['users'].add(principal)

    for r in roles:
      info = roles[r]
      info['groups'] = [u'{g} (id={g.id})'.format(g=g)
                        for g in sorted(info['groups'], key=lambda g: g.name)]
      users = sorted(
        info['users'],
        key=lambda u: (u.last_name.lower(), u.first_name.lower())
      )
      info['users'] = [u'{u} (id={u.id}, email="{u.email}")'.format(u=u)
                       for u in users]

    jinja_env = current_app.jinja_env
    jinja_env.filters.update(self.jinja_env.filters)
    template = jinja_env.get_or_select_template(
      'debug_panels/security_info_panel.html')

    return template.render(context)



"""
Starting work on a plugin system. This will probably be refactored heavily
in the future.
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import



# coding=utf-8
"""
Login-related views (login / logout / password reminder / ...).

Notes:
- Uses code copy/pasted (and modified) from Flask-Security
"""
from __future__ import absolute_import, division, print_function

import random
import string
from datetime import datetime
from urlparse import urljoin, urlparse

from flask import (current_app, flash, redirect, render_template, request,
                   url_for)
from flask.globals import _request_ctx_stack
from flask.json import jsonify
from flask_login import (login_user, logout_user, user_logged_in,
                         user_logged_out)
from flask_mail import Message
from flask_security.utils import md5
from itsdangerous import BadSignature, SignatureExpired, URLSafeTimedSerializer
from sqlalchemy import sql
from sqlalchemy.orm.exc import NoResultFound

from abilian.core.extensions import csrf, db
from abilian.core.models.subjects import User
from abilian.i18n import _, render_template_i18n
from abilian.services.security import Anonymous
from abilian.web.blueprints import Blueprint

from .models import LoginSession

__all__ = []

login = Blueprint("login", __name__,
                  url_prefix="/user",
                  allowed_roles=Anonymous,
                  template_folder='templates')
route = login.route


#
# Login / Logout
#
@route("/login")
def login_form():
  """Display the login form."""
  next_url = get_redirect_target()
  return render_template("login/login.html", next_url=next_url)


def do_login(form):
  email = form.get('email', "").lower()
  password = form.get('password')
  next_url = form.get('next', u'')
  res = dict(username=email, email=email, next_url=next_url)

  if not email or not password:
    res['error'] = _(u"You must provide your email and password.")
    res['code'] = 401
    return res

  try:
    user = User.query\
        .filter(sql.func.lower(User.email) == email,
                User.can_login == True)\
        .one()
  except NoResultFound:
    res['error'] = _(u"Sorry, we couldn't find an account for "
                     u"email '{email}'.").format(email=email)
    res['code'] = 401
    return res

  if user and not user.authenticate(password):
    res['error'] = _(u"Sorry, wrong password.")
    res['code'] = 401
    return res

  # Login successful
  login_user(user)
  res['user'] = user
  res['email'] = user.email
  return res


@csrf.exempt
@route("/login", methods=['POST'])
def login_post():
  res = do_login(request.form)
  if 'error' in res:
    code = res.pop('code')
    flash(res['error'], 'error')
    return (render_template("login/login.html", **res), code)

  return redirect_back(url=request.url_root)


@csrf.exempt
@route("/api/login", methods=['POST'])
def login_json():
  res = do_login(request.json)
  code = None

  if 'error' in res:
    code = res.pop('code')
  else:
    user = res.pop('user')
    res['fullname'] = user.name

  response = jsonify(**res)
  if code:
    response.status_code = code
  return response


@csrf.exempt
@route("/logout", methods=['GET', 'POST'])
def logout():
  logout_user()
  return redirect(request.url_root)


@csrf.exempt
@route("/api/logout", methods=['POST'])
def logout_json():
  logout_user()
  return  u'Logged out, ok.', 200


#
# Code to deal with forgotten passwords or initial password generation
# Note: most of this code is stolen (copy/pasted) from Flask-Security, with a
# few modifications.
#
@route("/forgotten_pw")
def forgotten_pw_form():
  return render_template("login/forgotten_password.html")


@route("/forgotten_pw", methods=['POST'])
@csrf.exempt
def forgotten_pw(new_user=False):
  """
  reset password for users who have already activated their accounts.
  """
  email = request.form.get('email', "").lower()

  action = request.form.get('action')
  if action == "cancel":
    return redirect(url_for("login.login_form"))

  if not email:
    flash(_(u"You must provide your email address."), 'error')
    return render_template("login/forgotten_password.html")

  try:
    user = User.query\
        .filter(sql.func.lower(User.email) == email,
                User.can_login == True)\
        .one()
  except NoResultFound:
    flash(_(u"Sorry, we couldn't find an account for "
            "email '{email}'.").format(email=email), 'error')
    return render_template("login/forgotten_password.html"), 401

  if user.can_login and not user.password:
    user.set_password(random_password())
    db.session.commit()

  send_reset_password_instructions(user)
  flash(_("Password reset instructions have been sent to your email address."),
        'info')

  return redirect(url_for("login.login_form"))


@route("/reset_password/<token>")
def reset_password(token):
  expired, invalid, user = reset_password_token_status(token)
  if invalid:
    flash(_(u"Invalid reset password token."), "error")
  elif expired:
    flash(_(u"Password reset expired"), "error")
  if invalid or expired:
    return redirect(url_for('login.forgotten_pw'))

  return render_template_i18n("login/password_reset.html")


@route("/reset_password/<token>", methods=['POST'])
@csrf.exempt
def reset_password_post(token):
  action = request.form.get('action')
  if action == "cancel":
    return redirect(request.url_root)

  expired, invalid, user = reset_password_token_status(token)

  if invalid or not user:
    flash(_(u"Invalid reset password token."), "error")
  elif expired:
    flash(_(u"Password reset expired"), "error")
  if invalid or expired or not user:
    return redirect(url_for('login.forgotten_pw'))

  password = request.form.get('password')

  if not password:
    flash(_(u"You must provide a password."), "error")
    return redirect(url_for("login.reset_password_post", token=token))

  # TODO: check entropy
  if len(password) < 7:
    flash(_(u"Your new password must be at least 8 characters long"), "error")
    return redirect(url_for("login.reset_password_post", token=token))

  if password.lower() == password:
    flash(_(u"Your new password must contain upper case and lower case "
            "letters"), "error")
    return redirect(url_for("login.reset_password_post", token=token))

  if not len([x for x in password if x.isdigit()]) > 0:
    flash(_(u"Your new password must contain at least one digit"), "error")
    return redirect(url_for("login.reset_password_post", token=token))

  user.set_password(password)
  db.session.commit()

  flash(_(u"Your password has been changed. "
          u"You can now login with your new password"),
        "success")

  return redirect(url_for("login.login_form", next_url=request.url_root))


def random_password():
  pw = []
  for i in range(0, 10):
    pw.append(random.choice(string.letters + string.digits))
  return pw


def get_serializer(name):
  """
  :type name: str
  """
  config = current_app.config
  secret_key = config.get('SECRET_KEY')
  salt = config.get('SECURITY_{}_SALT'.format(name.upper()))
  return URLSafeTimedSerializer(secret_key=secret_key, salt=salt)


def send_reset_password_instructions(user):
  """Sends the reset password instructions email for the specified user.

  :param user: The user to send the instructions to
  :type user: User
  """
  token = generate_reset_password_token(user)
  url = url_for('login.reset_password', token=token)
  reset_link = request.url_root[:-1] + url

  subject = _(u"Password reset instruction for {site_name}"
              ).format(site_name=current_app.config.get('SITE_NAME'))
  mail_template = 'password_reset_instructions'
  send_mail(subject, user.email, mail_template, user=user, reset_link=reset_link)


def generate_reset_password_token(user):
  """Generates a unique reset password token for the specified user.

  :param user: The user to work with
  :type user: User
  """
  data = [str(user.id), md5(user.password)]
  return get_serializer("reset").dumps(data)


def reset_password_token_status(token):
  """Returns the expired status, invalid status, and user of a password reset
  token. For example::

      expired, invalid, user = reset_password_token_status('...')

  :param token: The password reset token
  :type token: str
  """
  return get_token_status(token, 'reset', 'RESET_PASSWORD')


def get_token_status(token, serializer_name, max_age=None):
  serializer = get_serializer(serializer_name)
  #max_age = get_max_age(max_age)
  user, data = None, None
  expired, invalid = False, False

  try:
    data = serializer.loads(token, max_age=max_age)
  except SignatureExpired:
    d, data = serializer.loads_unsafe(token)
    expired = True
  except BadSignature:
    invalid = True

  if data:
    user = User.query.get(data[0])

  expired = expired and (user is not None)
  return expired, invalid, user


def send_mail(subject, recipient, template, **context):
  """Send an email using the Flask-Mail extension.

  :param subject: Email subject
  :param recipient: Email recipient
  :param template: The name of the email template
  :param context: The context to render the template with
  """

  config = current_app.config
  sender = config['MAIL_SENDER']
  msg = Message(subject,
                sender=sender,
                recipients=[recipient])

  ctx = ('login/email', template)
  msg.body = render_template_i18n('%s/%s.txt' % ctx, **context)
  #msg.html = render_template('%s/%s.html' % ctx, **context)

  mail = current_app.extensions.get('mail')
  current_app.logger.debug("Sending mail...")
  mail.send(msg)


#
# Logging
#
def log_session_start(app, user):
  session = LoginSession.new()
  db.session.add(session)
  db.session.commit()


def log_session_end(app, user):
  session = LoginSession.query.get_active_for(user)
  if session:
    session.ended_at = datetime.utcnow()
    db.session.commit()


user_logged_in.connect(log_session_start)
user_logged_out.connect(log_session_end)


# login redirect utilities
#  from http://flask.pocoo.org/snippets/62/
def is_safe_url(target):
  ref_url = urlparse(request.host_url)
  test_url = urlparse(urljoin(request.url_root, target))
  return test_url.scheme in ('http', 'https') and \
      ref_url.netloc == test_url.netloc


def check_for_redirect(target):
  target = urljoin(request.url_root, target)
  url = urlparse(target)
  reqctx = _request_ctx_stack.top
  try:
    endpoint, ignored = reqctx.url_adapter.match(url.path, 'GET')
    if '.' in endpoint and endpoint.rsplit('.', 1)[0] == 'login':
      # don't redirect to any login view after successful login
      return None
  except Exception as e:
    # exceptions may happen if route is not found for example
    pass

  return target


def get_redirect_target():
  for target in (request.values.get('next'), request.referrer):
    if not target:
      continue
    if is_safe_url(target):
      return check_for_redirect(target)


def redirect_back(endpoint=None, url=None, **values):
  target = request.form.get('next')
  if not target or not is_safe_url(target):
    if endpoint:
      target = url_for(endpoint, **values)
    else:
      target = url
  return redirect(check_for_redirect(target))



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from abilian.core.models import comment as comments
from abilian.web import url_for

from .forms import CommentForm
from .views import bp as blueprint
from .views import COMMENT_BUTTON


class CommentExtension(object):
  """
  API for comments, installed as an application extension.

  It is also available in templates as `comments`.
  """
  def __init__(self, app):
    app.extensions['comments'] = self
    app.add_template_global(self, 'comments')
    app.register_blueprint(blueprint)

  def is_commentable(self, obj):
    return comments.is_commentable(obj)

  def for_entity(self, obj, check_commentable=False):
    return comments.for_entity(obj, check_commentable=check_commentable)

  def has_comments(self, obj):
    return bool(comments.for_entity(obj, check_commentable=True))

  def count(self, obj):
    return len(comments.for_entity(obj, check_commentable=True))

  def get_form_context(self, obj):
    """
    Return a dict: form instance, action button, submit url...
    Used by macro m_comment_form(entity)
    """
    ctx = {}
    ctx['url'] = url_for('comments.create', entity_id=obj.id)
    ctx['form'] = CommentForm()
    ctx['buttons'] = [COMMENT_BUTTON]
    return ctx



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from wtforms.fields import TextAreaField

from abilian.core.models.comment import Comment
from abilian.i18n import _l
from abilian.web.forms import Form
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required
from abilian.web.forms.widgets import TextArea


class CommentForm(Form):

  body = TextAreaField(
    label=_l(u'Comment'),
    validators=[required()],
    filters=(strip,),
    widget=TextArea(rows=5, resizeable='vertical'),
  )

  class Meta:
    model = Comment
    include_primary_keys = True
    assign_required = False # for 'id': allow None, for new records



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import pygeoip
from flask import render_template

from abilian.i18n import _
from abilian.services.auth.models import LoginSession

from ..panel import AdminPanel

DATA_FILES = (
  '/usr/share/GeoIP/GeoIP.dat',
  '/usr/share/GeoIP/GeoIPv6.dat',
)

class LoginSessionsPanel(AdminPanel):
  id = "login_sessions"
  label = "Session log"
  icon = 'log-in'

  def get(self):
    geoips = []
    for filename in DATA_FILES:
      try:
        geoips.append(pygeoip.GeoIP(filename))
      except (pygeoip.GeoIPError, IOError):
        pass

    sessions = LoginSession.query.order_by(LoginSession.id.desc()).limit(50).all()
    unknown_country = _(u'Country unknown')

    for session in sessions:
      country = unknown_country

      if geoips and session.ip_address:
        ip_address = session.ip_address
        multiple = ip_address.split(',')
        if multiple:
          # only use last ip in the list, most likely the public address
          ip_address = multiple[-1]
        for g in geoips:
          try:
            country = g.country_name_by_addr(ip_address)
          except pygeoip.GeoIPError:
            continue

          if country:
            break
          else:
            country = unknown_country

        session.country = country

    return render_template("admin/login_sessions.html", **locals())



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import abc

import sqlalchemy as sa
from sqlalchemy import Column, ForeignKey, Integer, UnicodeText
from sqlalchemy.orm import backref, relationship

from abilian.core.entities import Entity
from abilian.services.security import CREATE, DELETE, WRITE, Anonymous, Owner

#: name of backref on target :class:`Entity` object
ATTRIBUTE = '__comments__'


class Commentable(object):
  """
  """
  __metaclass__ = abc.ABCMeta


def register(cls):
  """
  Register an :class:`Entity` as a commentable class.

  Can be used as a class decorator:

  .. code-block:: python

      @comment.register
      class MyContent(Entity):
          ...
  """
  if not issubclass(cls, Entity):
    raise ValueError('Class must be a subclass of abilian.core.entities.Entity')

  Commentable.register(cls)
  return cls


def is_commentable(obj):
  """
  :param obj: a class or instance
  """
  if isinstance(obj, type):
    return issubclass(obj, Commentable)

  if not isinstance(obj, Commentable):
    return False

  if obj.id is None:
    return False

  return True


def for_entity(obj, check_commentable=False):
  """
  Return comments on an entity.
  """
  if check_commentable and not is_commentable(obj):
    return []

  return getattr(obj, ATTRIBUTE)


class Comment(Entity):
  """
  A Comment related to an :class:`Entity`.
  """
  __default_permissions__ = {
    WRITE: {Owner},
    DELETE: {Owner},
    CREATE: {Anonymous},
  }

  @sa.ext.declarative.declared_attr
  def __mapper_args__(cls):
    # we cannot use super(Comment, cls): declared_attr happens during class
    # construction. super(cls, cls) could work; as long as `cls` is not a
    # subclass of `Comment`: it would enter into an infinite loop.
    #
    # Entity.__mapper_args__ calls the descriptor with 'Entity', not `cls`.
    args = Entity.__dict__['__mapper_args__'].fget(cls)
    args['order_by'] = cls.created_at
    return args

  entity_id = Column(Integer, ForeignKey(Entity.id), nullable=False)

  #: Commented entity
  entity = relationship(
    Entity,
    lazy='immediate',
    foreign_keys=[entity_id],
    backref=backref(ATTRIBUTE,
                    lazy='select',
                    order_by='Comment.created_at',
                    cascade="all, delete-orphan",
                  )
  )

  #: comment's main content
  body = Column(
    UnicodeText(),
    sa.CheckConstraint("trim(body) != ''"),
    nullable=False)

  @property
  def history(self):
    return self.meta.get('abilian.core.models.comment', {}).get('history', [])

  def __repr__(self):
    class_ = self.__class__
    mod_ = class_.__module__
    classname = class_.__name__
    return '<{}.{} instance at 0x{:x} entity id={!r}'\
        .format(mod_, classname, id(self), self.entity_id)



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from babel import Locale
from flask import _app_ctx_stack, _request_ctx_stack
from flask_babel import get_locale
from jinja2 import DictLoader

from abilian import i18n
from abilian.testing import BaseTestCase


class I18NTestCase(BaseTestCase):

  def test_set_locale(self):
    en = Locale('en')
    fr = Locale('fr')
    with self.app.test_request_context('/'):
      assert get_locale() == en

      with i18n.set_locale('fr') as new_locale:
        assert get_locale() == fr
        assert isinstance(new_locale, Locale)
        assert new_locale == fr
      assert get_locale() == en

      with i18n.set_locale(fr):
        assert get_locale() == fr
        with i18n.set_locale(en):
          assert get_locale() == en
        assert get_locale() == fr
      assert get_locale() == en

    # no request context: no locale set
    app_ctx = _app_ctx_stack.top
    self._ctx.pop()
    self._ctx = None
    app_ctx.push()
    with i18n.set_locale(fr):
      assert get_locale() is None


  def test_ensure_request_context(self):
    en = Locale('en')
    fr = Locale('fr')

    # by default, a request context is present is test case setup.  first let's
    # show no new request context is set up and that set_locale works as
    # advertised
    app_ctx = _app_ctx_stack.top
    current_ctx = _request_ctx_stack.top
    assert current_ctx is not None
    assert get_locale() == en

    with i18n.ensure_request_context():
      assert current_ctx == _request_ctx_stack.top

      with i18n.set_locale(fr):
        assert get_locale() == fr

    self._ctx.pop()
    self._ctx = None
    assert _app_ctx_stack.top is None
    assert _request_ctx_stack.top is None
    assert get_locale() is None

    app_ctx.push()
    # ensure set_locale() doesn't set up a request context
    with i18n.set_locale(fr):
      assert _request_ctx_stack.top is None
      assert get_locale() is None

    with i18n.ensure_request_context():
      assert _request_ctx_stack.top is not None
      assert current_ctx != _request_ctx_stack.top
      assert get_locale() == en

      with i18n.set_locale(fr):
        assert get_locale() == fr

    assert _request_ctx_stack.top is None
    assert get_locale() is None

    # test chaining
    with i18n.ensure_request_context(), i18n.set_locale(fr):
      assert get_locale() == fr

  def test_get_template_i18n(self):
    template_path = '/myfile.txt'
    en = Locale('en')
    result = i18n.get_template_i18n(template_path, locale=en)
    self.assertIn('/myfile.en.txt', result)
    self.assertIn('/myfile.txt', result)

    en = Locale('en_US')
    result = i18n.get_template_i18n(template_path, locale=en)
    self.assertIn('/myfile.en_US.txt', result)
    self.assertIn('/myfile.txt', result)

    with i18n.set_locale('fr'):
        result = i18n.get_template_i18n(template_path, get_locale())
        self.assertIn('/myfile.fr.txt', result)
        self.assertIn('/myfile.txt', result)

  def test_render_template_i18n(self):
    loader = DictLoader({u'tmpl.txt': u'default ({{ locale }})',
                         u'tmpl.en.txt': u'en locale ({{ locale }})',
                         u'tmpl.fr.txt': u'fr locale ({{ locale }})',})
    app_loader = self.app.jinja_loader
    self.app.jinja_loader = loader
    render = i18n.render_template_i18n
    try:
      assert render('tmpl.txt', locale='fr') == u'fr locale (fr)'
      assert render('tmpl.txt', locale='en') == u'en locale (en)'
      assert render('tmpl.txt', locale='de') == u'default (de)'
    finally:
      self.app.jinja_loader = app_loader

  def test_default_country(self):
    assert 'DEFAULT_COUNTRY' in self.app.config
    assert self.app.config['DEFAULT_COUNTRY'] is None
    assert i18n.default_country() is None
    assert i18n.country_choices()[0][0] == u'AF'

    self.app.config['DEFAULT_COUNTRY'] = u'FR'
    assert i18n.default_country() == u'FR'
    assert i18n.country_choices()[0][0] == u'FR'
    assert i18n.country_choices(default_country_first=False)[0][0] == u'AF'



# coding=utf-8
"""
Front-end for a CRM app.

This should eventually allow implementing very custom CRM-style application.
"""
from __future__ import absolute_import, division, print_function

import copy
import logging
import re
from collections import OrderedDict

import sqlalchemy as sa
from flask import (Blueprint, current_app, g, jsonify, redirect,
                   render_template, request, session, url_for)
from flask_login import current_user
from sqlalchemy import func, orm
from sqlalchemy.sql.expression import asc, desc, nullsfirst, nullslast
from werkzeug.exceptions import BadRequest

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.i18n import _l
from abilian.services import audit_service
from abilian.services.security import READ  # noqa
from abilian.services.vocabularies.models import BaseVocabulary

from . import search
from .action import (Action, ActionDropDown, ActionGroup, ActionGroupItem,
                     Endpoint, FAIcon, actions)
from .forms.widgets import (AjaxMainTableView, Panel, RelatedTableView, Row,
                            SingleView)
from .nav import BreadcrumbItem
from .views import (JSONView, JSONWhooshSearch, ObjectCreate, ObjectDelete,
                    ObjectEdit, ObjectView, default_view)

logger = logging.getLogger(__name__)


try:
  # sqlalchemy 0.8
  _DateAffinity = sa.types._DateAffinity
except AttributeError:
  _DateAffinity = sa.sql.sqltypes._DateAffinity

class ModuleAction(Action):
  """
  Base action class for :class:`Module` actions.

  Basic condition is simple: :attr:`category` must match the string
  `'module:{module.endpoint}'`
  """
  def __init__(self, module, group, name, *args, **kwargs):
    self.group = group
    super(ModuleAction, self).__init__(module.action_category, name,
                                       *args, **kwargs)

  def pre_condition(self, context):
    module = actions.context.get('module')
    if not module:
      return False

    return self.category == module.action_category


class ModuleActionGroup(ModuleAction, ActionGroup):
  template_string = ActionGroup.template_string


class ModuleActionDropDown(ModuleAction, ActionDropDown):
  template_string = ActionDropDown.template_string


class ModuleActionGroupItem(ModuleAction, ActionGroupItem):
  pass


def add_to_recent_items(entity, type='ignored'):
  if not isinstance(entity, Entity):
    return
  object_type = entity.object_type
  url = current_app.default_view.url_for(entity)
  if not hasattr(g, 'recent_items'):
    g.recent_items = []
  g.recent_items.insert(0, dict(type=object_type, name=entity.name, url=url))
  s = set()
  l = []
  for item in g.recent_items:
    if item['url'] in s:
      continue
    s.add(item['url'])
    l.append(item)
  if len(l) > 5:
    del l[5:]
  session['recent_items'] = g.recent_items = l


def expose(url='/', methods=('GET',)):
  """
  Use this decorator to expose views in your view classes.

  `url`
      Relative URL for the view
  `methods`
      Allowed HTTP methods. By default only GET is allowed.
  """

  def wrap(f):
    if not hasattr(f, '_urls'):
      f._urls = []
    f._urls.append((url, methods))
    return f

  return wrap


def labelize(s):
  return " ".join([w.capitalize() for w in s.split("_")])


def make_single_view(form, **options):
  panels = []
  for gr in form._groups.items():
    panel = Panel(gr[0], *[Row(x) for x in gr[1]])
    panels.append(panel)
  return SingleView(form, *panels, **options)


class ModuleView(object):
  """
  Mixin for module base views.

  Provide :attr:`module`.
  """
  #: :class:`Module` instance
  module = None

  def __init__(self, module, *args, **kwargs):
    self.module = module
    super(ModuleView, self).__init__(*args, **kwargs)


class BaseEntityView(ModuleView):
  pk = 'entity_id'

  def init_object(self, args, kwargs):
    args, kwargs = super(BaseEntityView, self).init_object(args, kwargs)

    # security check
    if self.obj and not self.check_access():
      # FIXME: will lead base views to return 404. Ok when permission is
      # READ, but when it's WRITE would 403 be more appropriate?
      self.obj = None

    return args, kwargs

  def breadcrumb(self):
    return BreadcrumbItem(label=self.obj.name or self.obj.id,
                          url=Endpoint('.entity_view', entity_id=self.obj.id))

  def prepare_args(self, args, kwargs):
    args, kwargs = super(BaseEntityView, self).prepare_args(args, kwargs)
    actions.context['module'] = self.module
    add_to_recent_items(self.obj)
    return args, kwargs

  def redirect_to_index(self):
    return redirect(self.module.url)

  @property
  def single_view(self):
    return make_single_view(self.form,
                            view_template=self.module.view_template,
                            view=self,
                            module=self.module,
                            **self.module.view_options)

  def check_access(self):
    return self._check_view_permission(self)

  def _check_view_permission(self, view):
    """
    :param view: a :class:`ObjectView` class or instance
    """
    security = current_app.services['security']
    return security.has_permission(current_user, view.permission, self.obj)

  @property
  def can_edit(self):
    return self._check_view_permission(self.module.edit_cls)

  @property
  def can_delete(self):
    return self._check_view_permission(self.module.delete_cls)

  @property
  def can_create(self):
    create_cls = self.module.create_cls
    permission = create_cls.permission
    cls_permissions = dict(self.Model.__default_permissions__)

    if self.permission in cls_permissions:
      security = current_app.services['security']
      return security.has_permission(current_user,
                                     create_cls.permission,
                                     obj=self.obj,
                                     roles=cls_permissions[permission])
    return False


EDIT_ACTION = Action(
  'module', 'object:view', title=_l(u'Edit'),
  button='default', condition=lambda ctx: ctx['view'].can_edit,
  icon=FAIcon('edit'),
  url=lambda ctx: url_for('.entity_edit',
                          **{ctx['view'].pk: ctx['view'].obj.id}),
)

DELETE_ACTION = Action(
  'module', 'object:view', title=_l(u'Delete'),
  button='danger', condition=lambda ctx: ctx['view'].can_delete,
  icon=FAIcon('trash fa-inverse'),
  url=lambda ctx: url_for('.entity_delete',
                          **{ctx['view'].pk: ctx['view'].obj.id}),
)
DELETE_ACTION.template = 'widgets/frontend_action_delete_confim.html'


class EntityView(BaseEntityView, ObjectView):
  mode = 'view'
  template = 'default/single_view.html'

  @property
  def object_actions(self):
    ctx = actions.context
    return [a for a in (EDIT_ACTION, DELETE_ACTION)
            if a.available(ctx)]

  @property
  def template_kwargs(self):
    module = self.module
    related_views = [v.render(self.obj) for v in module.related_views]
    rendered_entity = self.single_view.render(self.obj, self.form,
                                              related_views=related_views)
    audit_entries = audit_service.entries_for(self.obj)

    return dict(rendered_entity=rendered_entity,
                related_views=related_views,
                audit_entries=audit_entries,
                show_new_comment_form=True,
                show_new_attachment_form=True,
                module=self.module)


class EntityEdit(BaseEntityView, ObjectEdit):
  template = 'default/single_view.html'
  mode = 'edit'

  @property
  def template_kwargs(self):
    rendered_entity = self.single_view.render_form(self.form)
    return dict(rendered_entity=rendered_entity,
                show_new_comment_form=False,
                show_new_attachment_form=False,
                module=self.module)


class EntityCreate(BaseEntityView, ObjectCreate):
  template = 'default/single_view.html'
  mode = 'create'

  prepare_args = ObjectCreate.prepare_args
  breadcrumb = ObjectCreate.breadcrumb

  def check_access(self):
    return self.can_create

  @property
  def template_kwargs(self):
    rendered_entity = self.single_view.render_form(self.form)
    return dict(rendered_entity=rendered_entity,
                for_new=True,
                module=self.module)


class EntityDelete(BaseEntityView, ObjectDelete):
  pass


class ListJson(ModuleView, JSONView):
  """
  JSON endpoint, for AJAX-backed table views.
  """
  def data(self, *args, **kwargs):
    echo = int(kwargs.get("sEcho", 0))
    length = int(kwargs.get("iDisplayLength", 10))
    start = int(kwargs.get("iDisplayStart", 0))
    end = start + length

    total_count = self.module.listing_query.count()
    q = self.module.list_query(request)
    count = q.count()
    q = self.module.ordered_query(request, q)

    entities = q.slice(start, end).all()
    table_view = AjaxMainTableView(
        columns=self.module.list_view_columns,
        name=self.module.managed_class.__name__.lower(),
        ajax_source=url_for('.list_json'))

    data = [table_view.render_line(e) for e in entities]
    result = {
        "sEcho": echo,
        "iTotalRecords": total_count,
        "iTotalDisplayRecords": count,
        "aaData": data,
    }
    return result


class ModuleMeta(type):
  """
  Module metaclass.

  Does some precalculations (like getting list of view methods from the
  class) to avoid calculating them for each view class instance.
  """

  def __init__(cls, classname, bases, fields):
    type.__init__(cls, classname, bases, fields)

    # Gather exposed views
    cls._urls = []
    cls._default_view = None

    for p in dir(cls):
      attr = getattr(cls, p)

      if hasattr(attr, '_urls'):
        # Collect methods
        for url, methods in attr._urls:
          cls._urls.append((url, p, methods))

          if url == '/':
            cls._default_view = p

            # Wrap views
            # setattr(cls, p, _wrap_view(attr))

class ModuleComponent(object):
  """
  A component that provide new functions for a :class:`Module`
  """
  name = None

  def __init__(self, name=None):
    if name is not None:
      self.name = name

    if self.name is None:
      raise ValueError('A module component must have a name')

  def init_module(self, module):
    self.module = module
    self.init()

  def init(self, *args, **kwargs):
    """
    Implements this in components
    """
    pass

  def get_actions(self):
    return []


class Module(object):
  __metaclass__ = ModuleMeta

  id = None
  endpoint = None
  label = None
  managed_class = None
  list_view = None
  list_view_columns = []
  single_view = None
  components = ()

  # class based views. If not provided will be automaticaly created from
  # EntityView etc defined above
  base_template = 'base.html'
  view_cls = EntityView
  edit_cls = EntityEdit
  create_cls = EntityCreate
  delete_cls = EntityDelete
  json_search_cls = JSONWhooshSearch

  # form_class. Used when view_cls/edit_cls are not provided
  edit_form_class = None
  view_form_class = None  # by default, same as edit_form_class

  url = None
  name = None
  view_new_save_and_add = False  # show 'save and add new' button in /new form
  static_folder = None
  view_template = None
  view_options = None
  related_views = []
  blueprint = None
  search_criterions = (search.TextSearchCriterion("name",
                                                  attributes=('name', 'nom')),)
  tableview_options = {}  # used mostly to change datatable search_label
  _urls = []

  def __init__(self):
    # If endpoint name is not provided, get it from the class name
    if self.endpoint is None:
      class_name = self.__class__.__name__
      if class_name.endswith('Module'):
        class_name = class_name[0:-len('Module')]
      self.endpoint = class_name.lower()

    if self.label is None:
      self.label = labelize(self.endpoint)

    if self.id is None:
      self.id = self.managed_class.__name__.lower()

    # If name is not provided, use capitalized endpoint name
    if self.name is None:
      self.name = self._prettify_name(self.__class__.__name__)

    if self.view_options is None:
      self.view_options = {}

    # self.single_view = make_single_view(self.edit_form_class,
    #                                     view_template=self.view_template,
    #                                     **self.view_options)
    if self.view_form_class is None:
      self.view_form_class = self.edit_form_class

    # init class based views
    kw = dict(Model=self.managed_class,
              pk='entity_id',
              module=self,
              base_template=self.base_template)
    self._setup_view("/<int:entity_id>",
                     'entity_view',
                     self.view_cls,
                     Form=self.view_form_class,
                     **kw)
    view_endpoint = self.endpoint + '.entity_view'

    self._setup_view("/<int:entity_id>/edit",
                     'entity_edit',
                     self.edit_cls,
                     Form=self.edit_form_class,
                     view_endpoint=view_endpoint,
                     **kw)

    self._setup_view("/new",
                     'entity_new',
                     self.create_cls,
                     Form=self.edit_form_class,
                     chain_create_allowed=self.view_new_save_and_add,
                     view_endpoint=view_endpoint,
                     **kw)

    self._setup_view("/<int:entity_id>/delete",
                     'entity_delete',
                     self.delete_cls,
                     Form=self.edit_form_class,
                     view_endpoint=view_endpoint,
                     **kw)

    self._setup_view("/json", 'list_json', ListJson, module=self)

    self._setup_view('/json_search', 'json_search', self.json_search_cls,
                     Model=self.managed_class)

    self.init_related_views()

    # copy criterions instances; without that they may be shared by subclasses
    self.search_criterions = tuple((copy.deepcopy(c)
                                    for c in self.search_criterions))
    for sc in self.search_criterions:
      sc.model = self.managed_class

    self.__components = {}
    for component in self.components:
      component.init_module(self)
      self.__components[component.name] = component

  def get_component(self, name):
    return self.__components.get(name)

  def _setup_view(self, url, attr, cls, *args, **kwargs):
    """
    Register class based views
    """
    view = cls.as_view(attr, *args, **kwargs)
    setattr(self, attr, view)
    self._urls.append((url, attr, view.methods))

  def init_related_views(self):
    related_views = []
    for view in self.related_views:
      if not isinstance(view, RelatedView):
        view = DefaultRelatedView(*view)
      related_views.append(view)
    self.related_views = related_views

  @property
  def action_category(self):
    return u'module:{}'.format(self.endpoint)

  def get_grouped_actions(self):
    items = actions.for_category(self.action_category)
    groups = OrderedDict()
    for action in items:
      groups.setdefault(action.group, []).append(action)

    return groups

  def register_actions(self):
    ACTIONS = [
      ModuleAction(self, u'entity', u'create',
                   title=_l(u'Create New'), icon=FAIcon('plus'),
                   endpoint=Endpoint(self.endpoint + '.entity_new'),
                   button='default',),
    ]
    for component in self.components:
      ACTIONS.extend(component.get_actions())

    actions.register(*ACTIONS)

  def create_blueprint(self, crud_app):
    """
    Create a Flask blueprint for this module.
    """
    # Store admin instance
    self.crud_app = crud_app
    self.app = crud_app.app

    # If url is not provided, generate it from endpoint name
    if self.url is None:
      self.url = '%s/%s' % (self.crud_app.url, self.endpoint)
    else:
      if not self.url.startswith('/'):
        self.url = '%s/%s' % (self.crud_app.url, self.url)

    # Create blueprint and register rules
    self.blueprint = Blueprint(self.endpoint, __name__,
                               url_prefix=self.url)

    for url, name, methods in self._urls:
      self.blueprint.add_url_rule(url,
                                  name,
                                  getattr(self, name),
                                  methods=methods)

    # run default_view decorator
    default_view(self.blueprint,
                 self.managed_class,
                 id_attr='entity_id')(self.entity_view)

    # delay registration of our breadcrumbs to when registered on app; thus
    # 'parents' blueprint can register theirs befores ours
    self.blueprint.record_once(self._setup_breadcrumb_preprocessors)

    return self.blueprint

  def _setup_breadcrumb_preprocessors(self, state):
    self.blueprint.url_value_preprocessor(self._add_breadcrumb)

  def _add_breadcrumb(self, endpoint, values):
    g.breadcrumb.append(BreadcrumbItem(label=self.label,
                        url=Endpoint('.list_view')))

  @property
  def base_query(self):
    """
    Return a query instance for :attr:`managed_class`.
    """
    return self.managed_class.query

  @property
  def read_query(self):
    """
    Return a query instance for :attr:`managed_class` filtering on `READ`
    permission
    """
    return self.base_query.with_permission(READ)

  @property
  def listing_query(self):
    """
    Like `read_query`, but can be made lightweight with only columns and joins
    of interest.

    `read_query` can be used with exports for example, with lot more columns
    (generallly it means more joins).
    """
    return self.base_query.with_permission(READ)

  def query(self, request):
    """
    Return filtered query based on request args
    """
    args = request.args
    search = args.get("sSearch", "").replace("%", "").lower()
    q = self.read_query.distinct()

    for crit in self.search_criterions:
      q = crit.filter(q, self, request, search)

    return q

  def list_query(self, request):
    """
    Return a filtered query based on request args, for listings.

    Like `query`, but subclasses can modify it to remove costly joined loads for
    example.
    """
    args = request.args
    search = args.get("sSearch", "").replace("%", "").lower()
    q = self.listing_query
    q = q.distinct()

    for crit in self.search_criterions:
      q = crit.filter(q, self, request, search)

    return q

  def ordered_query(self, request, query=None):
    """ Order query according to request args.

    If query is None, the query is generated according to request args with
    self.query(request)
    """
    if query is None:
      query = self.query(request)

    engine = query.session.get_bind(self.managed_class.__mapper__)
    args = request.args
    sort_col = int(args.get("iSortCol_0", 1))
    sort_dir = args.get("sSortDir_0", "asc")
    sort_col_def = self.list_view_columns[sort_col]
    sort_col_name = sort_col_def['name']
    rel_sort_names = sort_col_def.get('sort_on', (sort_col_name,))
    sort_cols = []

    for rel_col in rel_sort_names:
      sort_col = getattr(self.managed_class, rel_col)
      if hasattr(sort_col, 'property') and isinstance(sort_col.property, orm.properties.RelationshipProperty):
        # this is a related model: find attribute to filter on
        query = query.outerjoin(sort_col_name, aliased=True)

        rel_model = sort_col.property.mapper.class_
        default_sort_name = 'name'
        if issubclass(rel_model, BaseVocabulary):
          default_sort_name = 'label'

        rel_sort_name = sort_col_def.get('relationship_sort_on', None)
        if rel_sort_name is None:
          rel_sort_name = sort_col_def.get('sort_on', default_sort_name)
        sort_col = getattr(rel_model, rel_sort_name)

      # XXX: Big hack, date are sorted in reverse order by default
      if isinstance(sort_col, _DateAffinity):
        sort_dir = 'asc' if sort_dir == 'desc' else 'desc'
      elif isinstance(sort_col, sa.types.String) or \
          hasattr(sort_col, 'property') and \
          isinstance(sort_col.property.columns[0].type, sa.types.String):
        sort_col = func.lower(sort_col)


      direction = desc if sort_dir == 'desc' else asc
      sort_col = direction(sort_col)

      # sqlite does not support 'NULLS FIRST|LAST' in ORDER BY clauses
      if engine.name != 'sqlite':
        nullsorder = nullslast if sort_dir == 'desc' else nullsfirst
        sort_col = nullsorder(sort_col)

      sort_cols.append(sort_col)

    query = query.order_by(*sort_cols)
    query.reset_joinpoint()
    return query

  #
  # Exposed views
  #
  @expose("/")
  def list_view(self):
    actions.context['module'] = self
    table_view = AjaxMainTableView(
      name=self.managed_class.__name__.lower(),
      columns=self.list_view_columns,
      ajax_source=url_for('.list_json'),
      search_criterions=self.search_criterions,
      options=self.tableview_options)
    rendered_table = table_view.render()

    ctx = dict(rendered_table=rendered_table,
               module=self,
               base_template=self.base_template
               )
    return render_template("default/list_view.html", **ctx)

  @expose("/json2")
  def list_json2(self):
    """
    Other JSON endpoint, this time used for filling select boxes dynamically.

    NB: not used currently.
    """
    args = request.args
    cls = self.managed_class

    q = args.get("q", "").replace("%", " ")
    if not q or len(q) < 2:
      raise BadRequest()

    query = db.session.query(cls.id, cls.name)
    query = query.filter(cls.name.ilike("%" + q + "%"))\
                 .distinct()\
                 .order_by(cls.name)\
                 .limit(50)
    all = query.all()

    result = {'results': [ { 'id': r[0], 'text': r[1]} for r in all ] }
    return jsonify(result)

  #
  # Utils
  #
  def is_current(self):
    return request.path.startswith(self.url)

  @staticmethod
  def _prettify_name(name):
    """
    Prettify class name by splitting name by capital characters.
    So, 'MySuperClass' will look like 'My Super Class'

    `name`
      String to prettify
    """
    return re.sub(r'(?<=.)([A-Z])', r' \1', name)


class RelatedView(object):
  """ A base class for related views
  """
  def render(self, entity):
    """
    Return a dict with keys 'label', 'attr_name', 'rendered', 'size',
    'show_empty', 'default_collapsed'
    """
    raise NotImplementedError


class DefaultRelatedView(RelatedView):
  """ Default view used by Module for items directly related to entity
  """

  def __init__(self, label, attr, column_names, options=None, show_empty=False):
    self.label = label
    self.attr = attr
    self.show_empty = show_empty
    self.column_names = column_names
    self.options = {}

    if options is not None:
      self.options.update(options)

  def render(self, entity):
    view = RelatedTableView(self.column_names, self.options)
    related_entities = getattr(entity, self.attr)
    return dict(label=self.label,
                attr_name=self.attr,
                rendered=view.render(related_entities, related_to=entity),
                show_empty=self.show_empty,
                size=len(related_entities))


# TODO: rename to CRMApp ?
class CRUDApp(object):

  modules = ()

  def __init__(self, app, modules=None, name=None):
    if name is None:
      name = self.__class__.__module__
      modules_signature = ','.join(str(module.id) for module in self.modules)
      name = name + '-' + modules_signature

    self.name = name
    self.app = app
    app.extensions[name] = self

    if modules:
      self.modules = modules

    for module in self.modules:
      self.add_module(module)

  def get_module(self, module_id):
    for m in self.modules:
      if m.id == module_id:
        return m

    return None

  def add_module(self, module):
    self.app.register_blueprint(self.create_blueprint(module))
    module.register_actions()

  def create_blueprint(self, module):
    return module.create_blueprint(self)



# coding=utf-8
"""
"""
from __future__ import absolute_import

import sqlalchemy as sa

from abilian.core.models.tag import Tag, entity_tag_tbl
from abilian.i18n import _

from ..search.criterion import BaseCriterion
from .extension import ENTITY_DEFAULT_NS_ATTR


class TagCriterion(BaseCriterion):
  """
  Filter entities with selected tag(s).
  """
  form_default_value = u''

  def __init__(self, *args, **kwargs):
    if len(args) == 0:
      kwargs.setdefault('name', 'tags')

    if len(args) < 2:
      kwargs.setdefault('label', _(u'Tags'))

    super(TagCriterion, self).__init__(*args, **kwargs)

  @BaseCriterion.model.setter
  def model(self, model):
    BaseCriterion.model.fset(self, model)
    self.ns = getattr(model, ENTITY_DEFAULT_NS_ATTR)

  @property
  def valid_tags(self):
    join_clause = entity_tag_tbl.join(
      self.model,
      self.model.id == entity_tag_tbl.c.entity_id,
    )
    model_tags = sa.sql.select([entity_tag_tbl.c.tag_id], from_obj=join_clause)
    return Tag.query.filter(Tag.ns == self.ns,
                            Tag.id.in_(model_tags))\
                    .all()

  def get_request_values(self, request):
    tag_ids = []

    for val in request.values.getlist(self.name):
      try:
        tag_ids.append(int(val))
      except ValueError:
        pass

    valid_tags = set(self.valid_tags)
    tags = []
    for tid in tag_ids:
      t = Tag.query.get(tid)
      if t in valid_tags:
        tags.append(t)

    return tags


  def filter(self, query, module, request, searched_text, *args, **kwargs):
    """
    """
    tags = self.get_request_values(request)
    if not tags:
      return query

    cond = sa.sql.exists(
      sa.sql.select(
        [1],
        sa.sql.and_(entity_tag_tbl.c.entity_id == self.model.id,
                    entity_tag_tbl.c.tag_id.in_(t.id for t in tags)),
      )
    )
    return query.filter(cond)

  @property
  def form_filter_type(self):
    return "select"

  @property
  def form_unset_value(self):
    return []

  @property
  def form_filter_args(self):
    # expected value: [list of selectable items, is multiple?]
    return [
      [(unicode(t.id), t.label) for t in self.valid_tags],
      True,
    ]



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import bleach
from wtforms.fields import BooleanField, IntegerField, StringField
from wtforms.widgets import HiddenInput

from abilian.i18n import _l
from abilian.web.forms import ModelForm
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required

ALLOWED_TAGS = [
    'b',
    'i',
    'del',
    's',
    'u',
    'small',
    'strong',
    'em',
]
ALLOWED_ATTRIBUTES = {}


class EditForm(ModelForm):
  label = StringField(_l(u'Label'), description=_l(u'allowed tags: %(tags)s', tags=u', '.join(ALLOWED_TAGS)),
                      filters=(strip,), validators=[required()])
  default = BooleanField(_l(u'Default'), default=False)
  active = BooleanField(_l(u'Active'), default=True)

  def validate_label(self, field):
    field.data = bleach.clean(field.data, tags=ALLOWED_TAGS,
                              attributes=ALLOWED_ATTRIBUTES, strip=True)


class ListEditForm(EditForm):
  id = IntegerField(widget=HiddenInput())
  position = IntegerField(widget=HiddenInput())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import unittest

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.models.subjects import Group, User
from abilian.testing import BaseTestCase

from . import (READ, WRITE, Admin, Anonymous, Authenticated, Creator,
               InheritSecurity, Owner, Permission, PermissionAssignment,
               Reader, Role, RoleAssignment, SecurityAudit, Writer, security)

TEST_EMAIL = u"joe@example.com"
TEST_PASSWORD = "tototiti"


def init_user():
  user = User(first_name=u"Joe", last_name=u"User", email=TEST_EMAIL,
              password=TEST_PASSWORD)
  db.session.add(user)
  db.session.flush()


class RoleTestCase(unittest.TestCase):

  def test_singleton(self):
    admin = Role('admin')
    other_admin = Role('admin')
    self.assertIs(admin, other_admin)
    self.assertEquals(id(admin), id(other_admin))

  def test_equality(self):
    admin = Role('admin')
    self.assertEquals(admin, 'admin')
    self.assertEquals(admin, u'admin')

  def test_ordering(self):
    roles = [Authenticated, Admin, Anonymous]
    roles.sort()
    assert roles == [Admin, Anonymous, Authenticated]

  def test_enumerate_assignables(self):
    assert Role.assignable_roles() == [Admin]


class IntegrationTestCase(BaseTestCase):

  def setUp(self):
    BaseTestCase.setUp(self)
    # init_user()
    security.init_app(self.app)
    security.start()

  def tearDown(self):
    security.stop()
    security.clear()
    BaseTestCase.tearDown(self)


class DummyModel(Entity):
  pass


class FolderishModel(Entity, InheritSecurity):
  pass


class SecurityTestCase(IntegrationTestCase):
  def test_anonymous_user(self):
    # anonymous user is not an SQLAlchemy instance and must be handled
    # specifically to avoid tracebacks
    anon = self.app.login_manager.anonymous_user()
    assert not security.has_role(anon, 'reader')
    assert security.has_role(anon, Anonymous)
    assert security.get_roles(anon) == [Anonymous]
    assert not security.has_permission(anon, 'read')

  def test_has_role_authenticated(self):
    anon = self.app.login_manager.anonymous_user()
    user = User(email=u"john@example.com", password="x")
    self.session.add(user)
    self.session.flush()
    assert not security.has_role(anon, Authenticated)
    assert security.has_role(user, Authenticated)

  def test_root_user(self):
    """ Root user always has any role, any permission
    """
    root = User.query.get(0)
    assert security.has_role(root, Admin)
    assert security.has_role(root, Anonymous)
    assert security.has_permission(root, 'manage')

    obj = DummyModel()
    self.session.add(obj)
    self.session.flush()
    assert security.has_role(root, Admin, obj)
    assert security.has_permission(root, 'manage', obj)

  def test_grant_basic_roles(self):
    user = User(email=u"john@example.com", password="x")
    self.session.add(user)
    self.session.flush()

    # everybody always has role Anonymous
    assert security.has_role(user, Anonymous)

    security.grant_role(user, Admin)
    assert security.has_role(user, Admin)
    assert security.get_roles(user) == [Admin]
    assert security.get_roles(user) == ['admin']
    assert security.get_principals(Admin) == [user]

    # clear roles cache for better coverage: has_permission uses
    # _fill_role_cache_batch(), get_roles uses _fill_role_cache()
    delattr(user, '__roles_cache__')
    assert security.has_permission(user, "read")
    assert security.has_permission(user, "write")
    assert security.has_permission(user, "manage")

    security.ungrant_role(user, "admin")
    assert not security.has_role(user, "admin")
    assert security.get_roles(user) == []
    assert security.get_principals(Admin) == []

    assert not security.has_permission(user, "read")
    assert not security.has_permission(user, "write")
    assert not security.has_permission(user, "manage")

  def test_grant_basic_roles_on_groups(self):
    user = User(email=u"john@example.com", password="x")
    group = Group(name=u"Test Group")
    user.groups.add(group)
    self.session.add(user)
    self.session.flush()

    security.grant_role(group, "admin")
    assert security.has_role(group, "admin")
    assert security.get_roles(group) == ['admin']
    assert security.get_principals(Admin) == [group]

    assert security.has_role(user, Admin)

    assert security.has_permission(user, "read")
    assert security.has_permission(user, "write")
    assert security.has_permission(user, "manage")

    security.ungrant_role(group, "admin")
    assert not security.has_role(group, "admin")
    assert security.get_roles(group) == []
    assert security.get_principals(Admin) == []

    assert not security.has_role(user, "admin")
    assert not security.has_permission(user, "read")
    assert not security.has_permission(user, "write")
    assert not security.has_permission(user, "manage")

  def test_grant_roles_on_objects(self):
    user = User(email=u"john@example.com", password=u"x")
    user2 = User(email=u"papa@example.com", password=u"p")
    group = Group(name=u"Test Group")
    user.groups.add(group)
    obj = DummyModel()
    self.session.add_all([user, user2, obj])
    self.session.flush()

    security.grant_role(user, 'global_role')
    security.grant_role(user, "reader", obj)
    assert security.has_role(user, "reader", obj)
    assert security.get_roles(user, obj) == ['reader']
    assert security.get_principals(Reader) == []
    assert security.get_principals(Reader, object=obj) == [user]

    assert security.has_permission(user, "read", obj)
    assert not security.has_permission(user, "write", obj)
    assert not security.has_permission(user, "manage", obj)

    # test get_roles "global": object roles should not appear
    assert security.get_roles(user) == ['global_role']

    # global role is valid on all object
    assert security.has_role(user, 'global_role', obj)

    security.ungrant_role(user, "reader", obj)
    assert not security.has_role(user, "reader", obj)
    assert security.get_roles(user, obj) == []
    assert security.has_role(user, 'global_role', obj)

    assert not security.has_permission(user, "read", obj)
    assert not security.has_permission(user, "write", obj)
    assert not security.has_permission(user, "manage", obj)

    # owner / creator roles
    assert security.get_principals(Owner, object=obj) == []
    assert security.get_principals(Creator, object=obj) == []
    old_owner = obj.owner
    old_creator = obj.creator
    obj.owner = user
    assert security.get_roles(user, obj) == [Owner]
    assert security.get_principals(Owner, object=obj) == [user]
    assert security.get_principals(Creator, object=obj) == []
    # if user2 has Admin role e gets the rights no matter Creator/Ownership
    security.grant_role(user2, Admin)
    assert security.has_role(user2, (Owner, Creator), obj)
    assert security.has_role(user, (Owner, Creator), obj)

    obj.owner = old_owner
    obj.creator = user
    assert security.get_roles(user, obj) == [Creator]
    assert security.get_principals(Owner, object=obj) == []
    assert security.get_principals(Creator, object=obj) == [user]
    obj.creator = old_creator

    # permissions through group membership
    security.grant_role(group, "manager", obj)
    assert security.has_role(group, "manager", obj)
    assert security.get_roles(group, obj) == ['manager']

    # group membership: user hasn't role set, but has permissions
    assert security.get_roles(user, obj, no_group_roles=True) == []
    assert security.has_permission(user, "read", obj)
    assert security.has_permission(user, "write", obj)
    assert security.has_permission(user, "manage", obj)

    group.members.remove(user)
    self.session.flush()
    assert not security.has_role(user, "manager", obj)
    assert security.get_roles(user, obj) == []
    assert not security.has_permission(user, "read", obj)
    assert not security.has_permission(user, "write", obj)
    assert not security.has_permission(user, "manage", obj)

    security.ungrant_role(group, "manager", obj)
    assert not security.has_role(group, "manager", obj)
    assert security.get_roles(group, obj) == []

    # when called on unmapped instance
    new_obj = DummyModel()
    assert not security.has_permission(user, READ, new_obj)

  def test_grant_roles_unique(self):
    user = User(email=u"john@example.com", password="x")
    obj = DummyModel()
    self.session.add_all([user, obj])
    self.session.flush()

    assert RoleAssignment.query.count() == 0

    security.grant_role(user, "manager", obj)
    self.session.flush()
    assert RoleAssignment.query.count() == 1

    security.grant_role(user, "manager", obj)
    self.session.flush()
    assert RoleAssignment.query.count() == 1

    security.grant_role(user, "reader", obj)
    self.session.flush()
    assert RoleAssignment.query.count() == 2

  def test_inherit(self):
    folder = FolderishModel()
    self.session.add(folder)
    self.session.flush()

    self.assertEquals(SecurityAudit.query.count(), 0)

    security.set_inherit_security(folder, False)
    self.session.flush()
    self.assertFalse(folder.inherit_security)
    self.assertEquals(SecurityAudit.query.count(), 1)

    security.set_inherit_security(folder, True)
    self.session.flush()
    self.assertTrue(folder.inherit_security)
    self.assertEquals(SecurityAudit.query.count(), 2)


  def test_add_list_delete_permissions(self):
    obj = DummyModel()
    assert security.get_permissions_assignments(obj) == {}
    self.session.add(obj)
    self.session.flush()

    security.add_permission(READ, Authenticated, obj)
    security.add_permission(READ, Owner, obj)
    security.add_permission(WRITE, Owner, obj)
    assert security.get_permissions_assignments(obj) == {
      READ: {Authenticated, Owner,},
      WRITE: {Owner,},
    }

    security.delete_permission(READ, Authenticated, obj)
    assert security.get_permissions_assignments(obj) == {
      READ: { Owner,},
      WRITE: {Owner,},
    }
    assert security.get_permissions_assignments(obj, READ) == {
      READ: { Owner,},
    }

    # do it twice: it should not crash
    security.add_permission(READ, Owner, obj)
    security.delete_permission(READ, Authenticated, obj)

    # set/get/delete global permission
    security.add_permission(READ, Writer)
    assert security.get_permissions_assignments() == {
      READ: {Writer,},
    }

  def test_has_permission_on_objects(self):
    has_permission = security.has_permission
    user = User(email=u"john@example.com", password=u"x")
    group = Group(name=u"Test Group")
    user.groups.add(group)
    obj = DummyModel(creator=user, owner=user)
    self.session.add_all([user, obj])
    self.session.flush()

    # global role provides permissions on any object
    security.grant_role(user, Reader)
    assert has_permission(user, READ, obj=obj)
    assert not has_permission(user, WRITE, obj=obj)

    security.grant_role(user, Writer, obj=obj)
    assert has_permission(user, WRITE, obj=obj)

    # permission assignment
    security.ungrant_role(user, Reader)
    security.ungrant_role(user, Writer, object=obj)
    security.grant_role(user, Authenticated)
    assert not has_permission(user, READ, obj=obj)
    assert not has_permission(user, WRITE, obj=obj)

    pa = PermissionAssignment(role=Authenticated, permission=READ, object=obj)
    self.session.add(pa)
    self.session.flush()
    assert has_permission(user, READ, obj=obj)
    assert not has_permission(user, WRITE, obj=obj)

    self.session.delete(pa)
    self.session.flush()
    assert not has_permission(user, READ, obj=obj)

    # Owner / Creator
    for role in (Owner, Creator):
      pa = PermissionAssignment(role=Owner, permission=READ, object=obj)
      self.session.add(pa)
      self.session.flush()
      assert has_permission(user, READ, obj=obj)

      self.session.delete(pa)
      self.session.flush()
      assert not has_permission(user, READ, obj=obj)


    # test when object is *not* in session (newly created objects have id=None
    # for instance)
    obj = DummyModel()
    assert security.has_role(user, Reader, object=obj) is False

  def test_has_permission_custom_roles(self):
    user = User(email=u"john@example.com", password="x")
    self.session.add(user)
    self.session.flush()

    role = Role('custom_role')
    permission = Permission('custom permission')
    assert not security.has_permission(user, permission, roles=role)
    security.grant_role(user, role)
    assert not security.has_permission(user, permission)
    assert security.has_permission(user, permission, roles=role)

    # Permission always granted if Anonymous role
    assert security.has_permission(user, permission, roles=Anonymous)

    # test convert legacy permission & implicit mapping
    security.grant_role(user, 'reader')
    assert security.has_permission(user, 'read')
    assert not security.has_permission(user, 'write')
    assert not security.has_permission(user, 'manage')

    security.grant_role(user, 'writer')
    assert security.has_permission(user, 'read')
    assert security.has_permission(user, 'write')
    assert not security.has_permission(user, 'manage')

    security.grant_role(user, 'manager')
    assert security.has_permission(user, 'read')
    assert security.has_permission(user, 'write')
    assert security.has_permission(user, 'manage')

  def test_query_entity_with_permission(self):
    get_filter = security.query_entity_with_permission
    user = User(email=u"john@example.com", password="x")
    self.session.add(user)

    obj_reader = DummyModel(name=u'reader')
    obj_writer = DummyModel(name=u'writer')
    obj_none = DummyModel(name=u'none')
    self.session.add_all([obj_reader, obj_writer, obj_none])

    self.session.add_all(
      [PermissionAssignment(role=Reader, permission=READ, object=obj_reader),
       PermissionAssignment(role=Writer, permission=WRITE, object=obj_writer)]
    )
    self.session.flush()

    # very unfiltered query returns all objects
    base_query = DummyModel.query
    assert set(base_query.all()) == {obj_reader, obj_writer, obj_none}

    # user has no roles: no objects returned at all
    assert base_query.filter(get_filter(READ, user=user)).all() == []
    assert base_query.filter(get_filter(WRITE, user=user)).all() == []

    # grant object specific roles
    security.grant_role(user, Reader, obj=obj_reader)
    security.grant_role(user, Writer, obj=obj_writer)
    self.session.flush()
    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]

    # Permission granted to anonymous: objects returned
    pa = PermissionAssignment(
      role=Anonymous,
      permission=WRITE,
      object=obj_reader)
    self.session.add(pa)

    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert set(base_query.filter(get_filter(WRITE, user=user)).all()) \
      == set([obj_reader, obj_writer])
    self.session.delete(pa)
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]

    # grant global roles
    security.ungrant_role(user, Reader, object=obj_reader)
    security.ungrant_role(user, Writer, object=obj_writer)
    security.grant_role(user, Reader)
    security.grant_role(user, Writer)
    self.session.flush()

    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]

    # admin role has all permissions
    # 1: local role
    security.ungrant_role(user, Reader)
    security.ungrant_role(user, Writer)
    security.grant_role(user, Admin, obj=obj_reader)
    security.grant_role(user, Admin, obj=obj_none)
    self.session.flush()

    assert set(base_query.filter(get_filter(READ, user=user)).all()) == \
      {obj_reader, obj_none}
    assert set(base_query.filter(get_filter(WRITE, user=user)).all()) == \
      {obj_reader, obj_none}

    # 2: global role
    security.ungrant_role(user, Admin, object=obj_reader)
    security.ungrant_role(user, Admin, object=obj_none)
    security.grant_role(user, Admin)
    self.session.flush()
    assert set(base_query.filter(get_filter(READ, user=user)).all()) == \
        {obj_reader, obj_writer, obj_none}

    # implicit role: Owner, Creator
    security.ungrant_role(user, Admin)
    assert base_query.filter(get_filter(READ, user=user)).all() == []
    assert base_query.filter(get_filter(WRITE, user=user)).all() == []

    obj_reader.creator = user
    obj_writer.owner = user
    self.session.add_all(
      [PermissionAssignment(role=Creator, permission=READ, object=obj_reader),
       PermissionAssignment(role=Owner, permission=WRITE, object=obj_writer)]
    )
    self.session.flush()

    assert base_query.filter(get_filter(READ, user=user)).all() == [obj_reader]
    assert base_query.filter(get_filter(WRITE, user=user)).all() == [obj_writer]


class PermissionNoSAWarnTestCase(IntegrationTestCase):

  SQLALCHEMY_WARNINGS_AS_ERROR = False

  def test_add_delete_permissions_expunged_obj(self):
    # weird case. In CreateObject based views, usually Entity is instanciated
    # and might be added to session if it has a relationship with an existing
    # object. `init_object` must do `session.expunge(obj)`. But entities will
    # have initialized default permissions during `after_attach`.
    #
    # At save time, the object is added again to session. The bug is that
    # without precaution we may create permissions assignment twice, because
    # assignments created in the first place are not yet again in session.(new,
    # dirty, deleted) and cannot be found with a filtered query on
    # PermissionAssignment because they have not been flushed yet.
    #
    security.add_permission(READ, Owner, None)
    obj = DummyModel()
    # override default permission at instance level
    obj.__default_permissions__ = frozenset({
      (READ, frozenset({Owner})),
    })
    self.session.add(obj) # core.entities._setup_default_permissions creates
                          # permissions
    security.add_permission(READ, Owner, obj) # no-op
    self.session.expunge(obj) # obj and its permissions are removed from session

    self.session.add(obj) # obj in session again. When
                          # _setup_default_permissions is called durint
                          # `after_flush`, previously created permission are not
                          # yet back in session. The cascading rule will add
                          # them just after (as of sqlalchemy 0.8, at least)


    # Finally the test! IntegrityError will be raised if we have done something
    # wrong (`Key (permission, role, object_id)=(..., ..., ...) already exists`)
    self.session.flush()



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

from abilian.core.signals import register_js_api
from abilian.web import url_for

from .criterion import BaseCriterion, TextSearchCriterion

__all__ = ['BaseCriterion', 'TextSearchCriterion']


def _do_register_js_api(sender):
  app = sender
  js_api = app.js_api.setdefault('search', {})
  # hack to avoid url_for escape '%'
  js_api['live'] = url_for('search.live', q='') + u'%QUERY'


def register_plugin(app):
  from .views import search
  app.register_blueprint(search)
  register_js_api.connect(_do_register_js_api)



# coding=utf-8
"""
Blueprint for views of dynamic images
"""
from __future__ import absolute_import, division, print_function

import colorsys
import hashlib
from datetime import datetime, timedelta

import pkg_resources
import sqlalchemy as sa
from flask import Blueprint, make_response, render_template, request
from pathlib import Path
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.models.blob import Blob
from abilian.core.models.subjects import User
from abilian.core.util import utc_dt
from abilian.services.image import CROP, RESIZE_MODES, get_size
from abilian.web.util import url_for

from .files import BaseFileDownload

blueprint = Blueprint('images', __name__, url_prefix='/images')
route = blueprint.route

DEFAULT_AVATAR = Path(
  pkg_resources.resource_filename('abilian.web',
                                  'resources/img/avatar-default.png')
)
DEFAULT_AVATAR_MD5 = hashlib.md5(DEFAULT_AVATAR.open('rb').read()).hexdigest()


class BaseImageView(BaseFileDownload):
  """
  """
  max_size = None

  def __init__(self, max_size=None, *args, **kwargs):
    super(BaseImageView, self).__init__(*args, **kwargs)
    # override class default value only if arg is specified in constructor. This
    # allows subclasses to easily override theses defaults.
    if max_size is not None:
      self.max_size = max_size

  def prepare_args(self, args, kwargs):
    args, kwargs = super(BaseImageView, self).prepare_args(args, kwargs)
    size = request.args.get('s', 0)
    try:
      size = int(size)
    except ValueError:
      raise BadRequest(
        'Invalid value for "s": {}. Not an integer.'.format(repr(size))
      )

    if self.max_size is not None:
      if size > self.max_size:
        raise BadRequest(
          'Size too large: {:d} (max: {:d})'.format(size, self.max_size)
        )

    kwargs['size'] = size

    resize_mode = request.args.get('m', CROP)
    if resize_mode not in RESIZE_MODES:
      resize_mode = CROP

    kwargs['mode'] = resize_mode
    return args, kwargs

  def make_response(self, image, size, mode, *args, **kwargs):
    """
    :param image: image as bytes
    :param s: requested maximum width/height size
    """
    from abilian.services.image import resize, get_format

    try:
      fmt = get_format(image)
    except IOError:
      #  not a known image file
      raise NotFound()

    self.content_type = u'image/png' if fmt == 'PNG' else u'image/jpeg'
    ext = u'.' + unicode(fmt.lower())

    filename = kwargs.get('filename')
    if not filename:
      filename = u'image'
    if not filename.lower().endswith(ext):
      filename += ext
    self.filename = filename

    if size:
      image = resize(image, size, size, mode=mode)
      if mode == CROP:
        assert get_size(image) == (size, size)
    else:
      image = image.read()

    return make_response(image)

  def get_filename(self, *args, **kwargs):
    return self.filename

  def get_content_type(self, *args, **kwargs):
    return self.content_type


class StaticImageView(BaseImageView):
  """
  View for static assets not served by static directory.

  Useful for default avatars for example.
  """
  expire_vary_arg = 'md5'

  def __init__(self, image, *args, **kwargs):
    """
    :param image: path to image file
    """
    BaseImageView.__init__(self, *args, **kwargs)
    self.image_path = Path(image)
    if not self.image_path.exists():
      p = unicode(self.image_path)
      raise ValueError('Invalid image path: {}'.format(repr(p)))

  def prepare_args(self, args, kwargs):
    kwargs['image'] = self.image_path.open('rb')
    kwargs['filename'] = self.image_path.name
    return BaseImageView.prepare_args(self, args, kwargs)


class BlobView(BaseImageView):
  """
  Default :attr:`expire_vary_arg` to `"md5"`.
  :attr:`set_expire` is set to `False` by default.
  """
  expire_vary_arg = 'md5'
  id_arg = 'object_id'

  def __init__(self, id_arg=None, *args, **kwargs):
    BaseImageView.__init__(self, *args, **kwargs)
    if id_arg is not None:
      self.id_arg = id_arg

  def prepare_args(self, args, kwargs):
    args, kwargs = BaseImageView.prepare_args(self, args, kwargs)
    b_id = kwargs.get(self.id_arg)
    try:
      b_id = int(b_id)
    except ValueError:
      raise BadRequest('Invalid image id: {}'.format(repr(b_id)))

    blob = Blob.query.get(b_id)
    if not blob:
      raise NotFound()

    meta = blob.meta
    filename = meta.get('filename',
                        meta.get('md5',
                                 unicode(blob.uuid)))
    kwargs['filename'] = filename
    kwargs['image'] = blob.file.open('rb')
    return args, kwargs


blob_image = BlobView.as_view('blob_image')
route("/files/<int:object_id>")(blob_image)


class UserMugshot(BaseImageView):
  expire_vary_arg = 'md5'

  def prepare_args(self, args, kwargs):
    args, kwargs = super(UserMugshot, self).prepare_args(args, kwargs)

    user_id = kwargs['user_id']
    user = User.query \
      .options(sa.orm.undefer(User.photo)) \
      .get(user_id)

    if user is None:
      raise NotFound()

    kwargs['user'] = user
    kwargs['image'] = user.photo
    return args, kwargs

  def make_response(self, user, image, size, *args, **kwargs):
    if image:
      #  user has set a photo
      return super(UserMugshot, self).make_response(image, size, *args, **kwargs)

    # render svg avatar
    if user.last_name:
      letter = user.last_name[0]
    elif user.first_name:
      letter = user.first_name[0]
    else:
      letter = u"?"
    letter = letter.upper()
    # generate bg color, pastel: sat=65% in hsl color space
    id_hash = hash((user.name + user.email).encode('utf-8'))
    hue = id_hash % 10
    hue = (hue * 36) / 360.0  # 10 colors: 360 / 10
    color = colorsys.hsv_to_rgb(hue, 0.65, 1.0)
    color = [int(x * 255) for x in color]
    color = u'rgb({0[0]}, {0[1]}, {0[2]})'.format(color)
    svg = render_template('default/avatar.svg',
                          color=color, letter=letter, size=size)
    response = make_response(svg)
    self.content_type = u'image/svg+xml'
    self.filename = u'avatar-{}.svg'.format(id_hash)
    return response


user_photo = UserMugshot.as_view('user_photo', set_expire=True, max_size=500)
route("/users/<int:user_id>")(user_photo)
route('/users/default')(StaticImageView.as_view('user_default',
                                                set_expire=True,
                                                image=DEFAULT_AVATAR, ))


def user_url_args(user, size):
  endpoint = 'images.user_default'
  kwargs = {'s': size,
            'md5': DEFAULT_AVATAR_MD5, }

  if not user.is_anonymous():
    endpoint = 'images.user_photo'
    kwargs['user_id'] = user.id
    content = (user.photo if user.photo
               else (user.name + user.email).encode('utf-8'))
    kwargs['md5'] = hashlib.md5(content).hexdigest()

  return endpoint, kwargs


def user_photo_url(user, size):
  """
  Return url to use for this user
  """
  endpoint, kwargs = user_url_args(user, size)
  return url_for(endpoint, **kwargs)
